import warnings
from typing import Optional, Union

import torch

from leaspy.exceptions import LeaspyIndividualParamsInputError, LeaspyModelInputError
from leaspy.io.data.dataset import Dataset
from leaspy.utils.functional import Exp, MatMul
from leaspy.utils.typing import DictParams, DictParamsTorch, FeatureType, KwargsType
from leaspy.utils.weighted_tensor import TensorOrWeightedTensor
from leaspy.variables.distributions import Normal
from leaspy.variables.specs import (
    Hyperparameter,
    IndividualLatentVariable,
    LatentVariableInitType,
    LinkedVariable,
    ModelParameter,
    NamedVariables,
    PopulationLatentVariable,
)
from leaspy.variables.state import State

from .mcmc_saem_compatible import McmcSaemCompatibleModel
from .obs_models import observation_model_factory

__all__ = ["TimeReparametrizedModel"]


class TimeReparametrizedModel(McmcSaemCompatibleModel):
    """
    Contains the common attributes & methods of the multivariate time-reparametrized models.

    Parameters
    ----------
    name : :obj:`str`
        Name of the model.
    source_dimension : Optional[:obj:`int`]
        Number of sources. Dimension of spatial components (default is None).
    **kwargs
        Hyperparameters for the model (including `obs_models`).

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        If inconsistent hyperparameters.
    """

    _xi_std = 0.5
    _tau_std = 5.0
    _noise_std = 0.1
    _sources_std = 1.0

    def __init__(
        self,
        name: str,
        source_dimension: Optional[int] = None,
        **kwargs,
    ):
        # TODO / WIP / TMP: dirty for now...
        # Should we:
        # - use factory of observation models instead? dataset -> ObservationModel
        # - or refact a bit `ObservationModel` structure? (lazy init of its variables...)
        # (cf. note in AbstractModel as well)
        dimension = kwargs.get("dimension", None)
        if "features" in kwargs:
            dimension = len(kwargs["features"])
        # source_dimension = kwargs.get("source_dimension", None)
        # if dimension == 1 and source_dimension not in {0, None}:
        #    raise LeaspyModelInputError(
        #        "You should not provide `source_dimension` != 0 for univariate model."
        #    )
        # self.source_dimension: Optional[int] = source_dimension
        observation_models = kwargs.get("obs_models", None)
        if observation_models is None:
            observation_models = (
                "gaussian-scalar" if dimension is None else "gaussian-diagonal"
            )
        if isinstance(observation_models, (list, tuple)):
            kwargs["obs_models"] = tuple(
                [
                    observation_model_factory(obs_model, **kwargs)
                    for obs_model in observation_models
                ]
            )
        elif isinstance(observation_models, dict):
            # Not really satisfied... Used for api load
            kwargs["obs_models"] = tuple(
                [
                    observation_model_factory(
                        observation_models["y"], dimension=dimension
                    )
                ]
            )
        else:
            kwargs["obs_models"] = (
                observation_model_factory(observation_models, dimension=dimension),
            )
        super().__init__(name, **kwargs)
        self._source_dimension = self._validate_source_dimension(source_dimension)

    @property
    def xi_std(self) -> torch.Tensor:
        """Return the standard deviation of xi as a tensor."""
        return torch.tensor([self._xi_std])

    @property
    def tau_std(self) -> torch.Tensor:
        """Return the standard deviation of tau as a tensor."""
        return torch.tensor([self._tau_std])

    @property
    def noise_std(self) -> torch.Tensor:
        """Return the standard deviation of the model as a tensor."""
        return torch.tensor(self._noise_std)

    @property
    def sources_std(self) -> float:
        """Return the standard deviation of sources as a float."""
        return self._sources_std

    @property
    def source_dimension(self) -> Optional[int]:
        """Return the number of the sources"""
        return self._source_dimension

    @source_dimension.setter
    def source_dimension(self, source_dimension: Optional[int] = None):
        """Set the dimensionality of the source space for the model."""
        self._source_dimension = self._validate_source_dimension(source_dimension)

    def _validate_source_dimension(self, source_dimension: Optional[int] = None) -> int:
        """
        Validate and sanitize the `source_dimension` parameter.

        Parameters
        ----------
        source_dimension : Optional[:obj:`int`], default=None
            The candidate source dimension to validate.

        Returns
        -------
        Optional[:obj:`int`]
            The validated source dimension value. Returns 0 if the model dimension is 1,
            otherwise returns the validated `source_dimension` or None if not provided.

        Raises
        ------
         :exc:`.LeaspyModelInputError`
            If `source_dimension` is not an integer, is negative, or exceeds the allowable range
            based on the model's dimension.
        """
        if self.dimension == 1:
            return 0
        if source_dimension is not None:
            if not isinstance(source_dimension, int):
                raise LeaspyModelInputError(
                    f"`source_dimension` must be an integer, not {type(source_dimension)}"
                )
            if source_dimension < 0:
                raise LeaspyModelInputError(
                    f"`source_dimension` must be >= 0, you provided {source_dimension}"
                )
            if self.dimension is not None and source_dimension > self.dimension - 1:
                raise LeaspyModelInputError(
                    f"Source dimension should be within [0, {self.dimension - 1}], "
                    f"you provided {source_dimension}"
                )
        return source_dimension

    @property
    def has_sources(self) -> bool:
        """
        Indicates whether the model includes sources.

        Returns
        -------
        :obj:`bool`
            True if `source_dimension` is a positive integer.
            False otherwise.
        """
        return (
            hasattr(self, "source_dimension")
            and isinstance(self.source_dimension, int)
            and self.source_dimension > 0
        )

    @staticmethod
    def time_reparametrization(
        *,
        t: TensorOrWeightedTensor[float],
        alpha: torch.Tensor,
        tau: torch.Tensor,
    ) -> TensorOrWeightedTensor[float]:
        """
        Tensorized time reparametrization formula.

        .. warning::
            Shapes of tensors must be compatible between them.

        Parameters
        ----------
        t : :class:`torch.Tensor`
            Timepoints to reparametrize
        alpha : :class:`torch.Tensor`
            Acceleration factors of individual(s)
        tau : :class:`torch.Tensor`
            Time-shift(s) of individual(s)

        Returns
        -------
        :class:`torch.Tensor`
            Reparametrized time of same shape as `timepoints`
        """
        return alpha * (t - tau)

    def get_variables_specs(self) -> NamedVariables:
        """
        Return the specifications of the variables (latent variables,
        derived variables, model 'parameters') that are part of the model.

        Returns
        -------
        NamedVariables :
            A dictionary-like object containing specifications for the variables
        """
        specifications = super().get_variables_specs()
        specifications.update(
            rt=LinkedVariable(self.time_reparametrization),
            # PRIORS
            tau_mean=ModelParameter.for_ind_mean("tau", shape=(1,)),
            tau_std=ModelParameter.for_ind_std("tau", shape=(1,)),
            xi_std=ModelParameter.for_ind_std("xi", shape=(1,)),
            # LATENT VARS
            xi=IndividualLatentVariable(Normal("xi_mean", "xi_std")),
            tau=IndividualLatentVariable(Normal("tau_mean", "tau_std")),
            # DERIVED VARS
            alpha=LinkedVariable(Exp("xi")),
        )
        if self.source_dimension >= 1:
            specifications.update(
                # PRIORS
                betas_mean=ModelParameter.for_pop_mean(
                    "betas",
                    shape=(self.dimension - 1, self.source_dimension),
                ),
                betas_std=Hyperparameter(0.01),
                sources_mean=Hyperparameter(torch.zeros((self.source_dimension,))),
                sources_std=Hyperparameter(1.0),
                # LATENT VARS
                betas=PopulationLatentVariable(
                    Normal("betas_mean", "betas_std"),
                    sampling_kws={"scale": 0.5},  # cf. GibbsSampler (for retro-compat)
                ),
                sources=IndividualLatentVariable(Normal("sources_mean", "sources_std")),
                # DERIVED VARS
                mixing_matrix=LinkedVariable(
                    MatMul("orthonormal_basis", "betas").then(torch.t)
                ),  # shape: (Ns, Nfts)
                space_shifts=LinkedVariable(
                    MatMul("sources", "mixing_matrix")
                ),  # shape: (Ni, Nfts)
            )

        return specifications

    def _validate_compatibility_of_dataset(
        self, dataset: Optional[Dataset] = None
    ) -> None:
        """
        Validate the compatibility of the provided dataset with the model's configuration.

        Parameters
        ----------
        dataset : Optional[Dataset], optional
            The dataset to validate against, by default None.

        Raises
        ------
        LeaspyModelInputError
            If `source_dimension` is provided but not an integer in the valid range
            [0, dataset.dimension - 1).
        """
        super()._validate_compatibility_of_dataset(dataset)
        if not dataset:
            return
        if self.source_dimension is None:
            self.source_dimension = int(dataset.dimension**0.5)
            warnings.warn(
                "You did not provide `source_dimension` hyperparameter for multivariate model, "
                f"setting it to ⌊√dimension⌋ = {self.source_dimension}."
            )
        elif not (
            isinstance(self.source_dimension, int)
            and 0 <= self.source_dimension < dataset.dimension
        ):
            raise LeaspyModelInputError(
                f"Sources dimension should be an integer in [0, dimension - 1[ "
                f"but you provided `source_dimension` = {self.source_dimension} "
                f"whereas `dimension` = {dataset.dimension}."
            )

    def _audit_individual_parameters(
        self, individual_parameters: DictParams
    ) -> KwargsType:
        """
        Validate and process individual parameter inputs for model compatibility.

        Parameters
        ----------
        individual_parameters : DictParams
            A dictionary mapping parameter names (strings) to their values,
            which can be scalars or array-like structures.

        Returns
        -------
        KwargsType
            A dictionary with the following keys:
            - "nb_inds": Number of individuals
            - "tensorized_ips": Dictionary of parameters converted to 2D tensors.
            - "tensorized_ips_gen": Generator yielding tensors for each individual,
            each with an added batch dimension.

        Raises
        ------
        LeaspyIndividualParamsInputError
            If the provided dictionary keys do not match the expected parameter names,
            or if the sizes of individual parameters are inconsistent,
            or if `sources` parameter does not meet array-like requirements.
        """
        from .utilities import is_array_like, tensorize_2D

        expected_parameters = set(["xi", "tau"] + int(self.has_sources) * ["sources"])
        given_parameters = set(individual_parameters.keys())
        symmetric_diff = expected_parameters.symmetric_difference(given_parameters)
        if len(symmetric_diff) > 0:
            raise LeaspyIndividualParamsInputError(
                f"Individual parameters dict provided {given_parameters} "
                f"is not compatible for {self.name} model. "
                f"The expected individual parameters are {expected_parameters}."
            )
        ips_is_array_like = {
            k: is_array_like(v) for k, v in individual_parameters.items()
        }
        ips_size = {
            k: len(v) if ips_is_array_like[k] else 1
            for k, v in individual_parameters.items()
        }
        if self.has_sources:
            if not ips_is_array_like["sources"]:
                raise LeaspyIndividualParamsInputError(
                    f"Sources must be an array_like but {individual_parameters['sources']} was provided."
                )
            tau_xi_scalars = all(ips_size[k] == 1 for k in ["tau", "xi"])
            if tau_xi_scalars and (ips_size["sources"] > 1):
                # is 'sources' not a nested array? (allowed iff tau & xi are scalars)
                if not is_array_like(individual_parameters["sources"][0]):
                    # then update sources size (1D vector representing only 1 individual)
                    ips_size["sources"] = 1
            # TODO? check source dimension compatibility?
        uniq_sizes = set(ips_size.values())
        if len(uniq_sizes) != 1:
            raise LeaspyIndividualParamsInputError(
                f"Individual parameters sizes are not compatible together. Sizes are {ips_size}."
            )
        # number of individuals present
        n_individual_parameters = uniq_sizes.pop()
        # properly choose unsqueezing dimension when tensorizing array_like (useful for sources)
        # [1,2] => [[1],[2]] (expected for 2 individuals / 1D sources)
        # [1,2] => [[1,2]] (expected for 1 individual / 2D sources)
        unsqueeze_dim = 0 if n_individual_parameters == 1 else -1
        # tensorized (2D) version of ips
        tensorized_individual_parameters = {
            name: tensorize_2D(value, unsqueeze_dim=unsqueeze_dim)
            for name, value in individual_parameters.items()
        }

        return {
            "nb_inds": n_individual_parameters,
            "tensorized_ips": tensorized_individual_parameters,
            "tensorized_ips_gen": (
                {
                    name: value[individual, :].unsqueeze(0)
                    for name, value in tensorized_individual_parameters.items()
                }
                for individual in range(n_individual_parameters)
            ),
        }

    def _load_hyperparameters(self, hyperparameters: KwargsType) -> None:
        """
        Updates all model hyperparameters from the provided dictionary.

        Parameters
        ----------
        hyperparameters : KwargsType
            Dictionary containing the hyperparameters to be loaded.
            Expected keys include:
            - "features": List or sequence of feature names
            - "dimension": Integer specifying the number of features
            - "source_dimension": Integer specifying the number of sources; must be in
            [0, dimension - 1].

        Raises
        ------
        LeaspyModelInputError
            If `dimension` does not match the length of `features`, or if `source_dimension`
            is not an integer within the valid range [0, dimension - 1].
        """
        if "features" in hyperparameters:
            self.features = hyperparameters["features"]

        if "dimension" in hyperparameters:
            if self.features and hyperparameters["dimension"] != len(self.features):
                raise LeaspyModelInputError(
                    f"Dimension provided ({hyperparameters['dimension']}) does not match "
                    f"features ({len(self.features)})"
                )
            self.dimension = hyperparameters["dimension"]

        if "source_dimension" in hyperparameters:
            if not (
                isinstance(hyperparameters["source_dimension"], int)
                and (hyperparameters["source_dimension"] >= 0)
                and (
                    self.dimension is None
                    or hyperparameters["source_dimension"] <= self.dimension - 1
                )
            ):
                raise LeaspyModelInputError(
                    f"Source dimension should be an integer in [0, dimension - 1], "
                    f"not {hyperparameters['source_dimension']}"
                )
            self.source_dimension = hyperparameters["source_dimension"]

    def put_individual_parameters(self, state: State, dataset: Dataset):
        """
        Initialize individual latent parameters in the given state if not already set.
        
        Parameters
        ----------
        state : State
            The current state object that holds all the variables
        dataset : Dataset
            Dataset used to initialize latent variables accordingly.
        """
        if not state.are_variables_set(("xi", "tau")):
            with state.auto_fork(None):
                state.put_individual_latent_variables(
                    LatentVariableInitType.PRIOR_SAMPLES,
                    n_individuals=dataset.n_individuals,
                )

    def to_dict(self, *, with_mixing_matrix: bool = True) -> KwargsType:
        """
        Export model object as dictionary ready for :term:`JSON` saving.

        Parameters
        ----------
        with_mixing_matrix : :obj:`bool` (default ``True``)
            Save the :term:`mixing matrix` in the exported file in its 'parameters' section.

            .. warning::
                It is not a real parameter and its value will be overwritten at model loading
                (orthonormal basis is recomputed from other "true" parameters and mixing matrix
                is then deduced from this orthonormal basis and the betas)!
                It was integrated historically because it is used for convenience in
                browser webtool and only there...

        Returns
        -------
        KwargsType :
            The object as a dictionary.
        """
        model_settings = super().to_dict()
        model_settings["source_dimension"] = self.source_dimension

        if with_mixing_matrix and self.source_dimension >= 1:
            # transposed compared to previous version
            model_settings["parameters"]["mixing_matrix"] = self.state[
                "mixing_matrix"
            ].tolist()

        return model_settings

    # TODO: unit tests? (functional tests covered by api.estimate)
    def compute_individual_ages_from_biomarker_values(
        self,
        value: Union[float, list[float]],
        individual_parameters: DictParams,
        feature: Optional[FeatureType] = None,
    ) -> torch.Tensor:
        """
        For one individual, compute age(s) at which the given features values
        are reached (given the subject's individual parameters).

        Consistency checks are done in the main :term:`API` layer.

        Parameters
        ----------
        value : scalar or array_like[scalar] (:obj:`list`, :obj:`tuple`, :class:`numpy.ndarray`)
            Contains the :term:`biomarker` value(s) of the subject.

        individual_parameters : :obj:`dict`
            Contains the individual parameters.
            Each individual parameter should be a scalar or array_like.

        feature : :obj:`str` (or None)
            Name of the considered :term:`biomarker`.

            .. note::
                Optional for :class:`.UnivariateModel`, compulsory
                for :class:`.MultivariateModel`.

        Returns
        -------
        :class:`torch.Tensor`
            Contains the subject's ages computed at the given values(s).
            Shape of tensor is ``(1, n_values)``.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If computation is tried on more than 1 individual.
        """
        # value, individual_parameters = self._get_tensorized_inputs(
        #     value, individual_parameters, skip_ips_checks=False
        # )
        # return self.compute_individual_ages_from_biomarker_values_tensorized(
        #     value, individual_parameters, feature
        # )
        raise NotImplementedError("This method is currently not implemented.")

    def compute_individual_ages_from_biomarker_values_tensorized(
        self,
        value: torch.Tensor,
        individual_parameters: DictParamsTorch,
        feature: Optional[FeatureType],
    ) -> torch.Tensor:
        """
        For one individual, compute age(s) at which the given features values are
        reached (given the subject's individual parameters), with tensorized inputs.

        Parameters
        ----------
        value : :class:`torch.Tensor` of shape ``(1, n_values)``
            Contains the :term:`biomarker` value(s) of the subject.

        individual_parameters : DictParamsTorch
            Contains the individual parameters.
            Each individual parameter should be a :class:`torch.Tensor`.

        feature : :obj:`str` (or None)
            Name of the considered :term:`biomarker`.

            .. note::
                Optional for :class:`.UnivariateModel`, compulsory
                for :class:`.MultivariateModel`.

        Returns
        -------
        :class:`torch.Tensor`
            Contains the subject's ages computed at the given values(s).
            Shape of tensor is ``(n_values, 1)``.
        """
        raise NotImplementedError("This method is currently not implemented.")



"""Define some useful type aliases for static type checks and better input understanding."""

from typing import Any

import torch

__all__ = [
    "KwargsType",
    "IDType",
    "ParamType",
    "FeatureType",
    "DictParams",
    "DictParamsTorch",
]

# Generic dictionary of keyword arguments
KwargsType = dict[str, Any]

# Type for identifier of individuals
IDType = str

# Type for parameters / variables (mostly in dictionary)
ParamType = str

# Type for feature names
FeatureType = str

DictParams = dict[ParamType, Any]
DictParamsTorch = dict[ParamType, torch.Tensor]



import warnings
from abc import abstractmethod

import numpy as np
import pandas as pd

from leaspy.exceptions import LeaspyDataInputError
from leaspy.utils.typing import IDType

from .individual_data import IndividualData

__all__ = ["AbstractDataframeDataReader"]


class AbstractDataframeDataReader:
    """
    Methods to convert :class:`pandas.DataFrame` to `Leaspy`-compliant data containers.

    Raises
    ------
    :exc:`.LeaspyDataInputError`
    """

    time_rounding_digits = 6

    def __init__(self):
        self.individuals: dict[IDType, IndividualData] = {}
        self.iter_to_idx: dict[int, IDType] = {}
        self.n_individuals: int = 0

    ######################################################
    #               COMMON METHODS
    ######################################################

    @staticmethod
    def _check_numeric_type(dtype) -> bool:
        """
        Check if the type of the pandas data is numeric or not

        Parameters
        ----------
        s: pandas.Series.dtype
            pandas type of the data

        Returns
        -------
        : :obj:`bool`
            True if the type is a numeric type
        """
        return pd.api.types.is_numeric_dtype(
            dtype
        ) and not pd.api.types.is_complex_dtype(dtype)

    @classmethod
    def _check_ID(cls, s: pd.Series) -> None:
        """
        Check requirements on subjects identifiers.

        Parameters
        ----------
        s: pd.Series
            Identifiers of the patients

        Raises
        ------
        :exc:`.LeaspyModelInputError` :
            - If the :s:`pd.Series` is not a string, integer or categories
            - If the :s:`pd.Series` contains Nan
            - If the :s:`pd.Series` is integer and contain negative values
            - If the :s:`pd.Series` is string and contain empty strings
        """
        # TODO? enforce strings? (for compatibility for downstream requirements especially in IndividualParameters)
        valid_dtypes = ["string", "integer", "categorical"]
        inferred_dtype = pd.api.types.infer_dtype(s)
        if inferred_dtype not in valid_dtypes:
            raise LeaspyDataInputError(
                "The `ID` column should identify individuals as string, integer or categories, "
                f"not {inferred_dtype} ({s.dtype})."
            )

        if s.isna().any():
            # NOTE: as soon as a np.nan or np.inf, inferred_dtype cannot be 'integer'
            # but custom pandas dtype can still contain pd.NA
            raise LeaspyDataInputError(
                f"The `ID` column should NOT contain any nan ({s.isna().sum()} found)."
            )

        if inferred_dtype == "integer":
            if (s < 0).any():
                raise LeaspyDataInputError(
                    "All `ID` should be >= 0 when subjects are identified as integers, "
                    "use string identifiers if you need more flexibility."
                )
        elif inferred_dtype == "string":
            if (s.str.len() == 0).any():
                raise LeaspyDataInputError(
                    "No `ID` should be empty when subjects are identified as strings."
                )

    def _clean_index(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Check requirements on subjects identifiers:
            - ID represents patient index,
            - TIME represents the "age" of the patient at visit for visit indexing.

        Parameters
        ----------
        df: pd.DataFrame
            The whole dataframe with patient information.

        Returns
        -------
        df: pd.DataFrame
            The whole dataframe with patient information with a clean, set index
        """

        df = df.copy(deep=True)

        # Check columns headers
        columns = df.columns.tolist()

        # Try to read the raw dataframe
        try:
            self._check_headers(columns)

        # If we do not find 'ID' and 'TIME' columns, check the Index
        except LeaspyDataInputError:
            df.reset_index(inplace=True)
            columns = df.columns.tolist()
            self._check_headers(columns)

        # Check patient ID common to every format
        self._check_ID(df["ID"])

        df = self._set_index(df)
        if not df.index.is_unique:
            # get lines number as well as ID & TIME for duplicates (original line numbers)
            df_dup = df[[]].reset_index().duplicated(keep=False)
            df_dup = df_dup[df_dup]
            raise LeaspyDataInputError(f"Some raw are duplicated:\n{df_dup}")

        return df

    def _clean_numeric_data(
        self, df: pd.DataFrame, drop_full_nan: bool, warn_empty_column: bool
    ) -> pd.DataFrame:
        """
        Dataframe with patient information should only contain numeric data. This method check that this is the
        case, clean nans and empty columns.

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information.

        drop_full_nan: bool
            If set to True, raw full of nan are dropped.

        warn_empty_column: bool
            If set to True, a warning is raise for columns full of nan.

        Returns
        -------
        df: pd.DataFrame
            Dataframe with clean numeric information.
        """
        # Check and clean numerical values
        types_nok = {
            ft: dtype
            for ft, dtype in df.dtypes.items()
            if not self._check_numeric_type(dtype)
        }
        if types_nok:
            raise LeaspyDataInputError(
                f"All columns should be of numerical type, which is not the case for {types_nok}."
            )

        try:
            # it is needed so to always use numpy.nan as nans even if pd.NA were used originally
            df = df.astype(float)
        except Exception as e:
            raise LeaspyDataInputError(
                "Cannot safely convert dataframe to float type."
            ) from e

        # warn if some columns are full of nans
        full_of_nans = df.isna().all(axis=0)
        full_of_nans = full_of_nans[full_of_nans].index.tolist()
        if warn_empty_column and full_of_nans:
            warnings.warn(f"These columns only contain nans: {full_of_nans}.")

        # check that no 'inf' are present in dataframe
        df_inf = np.isinf(df)  # numpy.nan are considered finite :)
        df_inf_rows_and_cols = (
            df.where(df_inf)
            .dropna(how="all", axis=0)
            .dropna(how="all", axis=1)
            .fillna("")
        )
        if len(df_inf_rows_and_cols) != 0:
            raise LeaspyDataInputError(
                f"Values may be nan but not infinite, double check your data:\n{df_inf_rows_and_cols}"
            )

        # Drop visits full of nans so to get a correct number of total visits
        if drop_full_nan:
            df = df.dropna(how="all")

        return df

    @abstractmethod
    def _check_headers(self, columns: list[str]) -> None:
        """
        Check mendatory dataframe headers

        Parameters
        ----------
        columns: List[str]
            Names of the columns headers of the dataframe that contains patients information
        """

    @abstractmethod
    def _set_index(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Set the index suited for the type of information contained in the dataframe

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information

        Returns
        -------
        df: pd.DataFrame
            Dataframe with the right index
        """

    @abstractmethod
    def _clean_dataframe(
        self, df: pd.DataFrame, *, drop_full_nan: bool, warn_empty_column: bool
    ) -> pd.DataFrame:
        """
        Clean the dataframe that contains patient information. This method depends on the data type that is analysed: repeated measures, events or both

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information.

        drop_full_nan: bool
            If set to True, raw full of nan are dropped.

        warn_empty_column: bool
            If set to True, a warning is raise for columns full of nan.


        Returns
        -------
        df: pd.DataFrame
            Dataframe with clean information.
        """

    @abstractmethod
    def _load_individuals_data(
        self, subj: IndividualData, df_subj: pd.DataFrame
    ) -> None:
        """
        Convert information stored in a dataframe to information stored into IndividualData

        Parameters
        ----------
        subj: IndividualData
            One patient with her/his information, potentially empty

        df_subj: pd.DataFrame
            One patient with her/his information
        """

    ######################################################
    #               MAIN METHOD
    ######################################################

    def read(
        self,
        df: pd.DataFrame,
        *,
        drop_full_nan: bool = True,
        sort_index: bool = False,
        warn_empty_column: bool = True,
    ) -> None:
        """
        The method that effectively reads the input dataframe (automatically called in __init__).

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            The dataframe to read.
        drop_full_nan : bool
            Should we drop rows full of nans? (except index)
        sort_index : bool
            Should we lexsort index?
            (Keep False as default so not to break many of the downstream tests that check order...)
        warn_empty_column : bool
            Should we warn when there are empty columns?
        """
        if not isinstance(df, pd.DataFrame):
            # TODO? accept series? (for univariate dataset, with index already set)
            raise LeaspyDataInputError(
                "Input should be a pandas.DataFrame not anything else."
            )

        df = df.copy(deep=True)  # No modification on the input dataframe !
        df = self._clean_index(df)
        df = self._clean_numeric_data(df, drop_full_nan, warn_empty_column)

        # Clean data
        df = self._clean_dataframe(
            df, drop_full_nan=drop_full_nan, warn_empty_column=warn_empty_column
        )

        # sort after duplicate checks and full of nans dropped
        if sort_index:
            df.sort_index(inplace=True)

        # Create individuals to store
        for idx_subj, df_subj in df.groupby(level="ID", sort=False):
            self.individuals[idx_subj] = IndividualData(idx_subj)
            self._load_individuals_data(self.individuals[idx_subj], df_subj)

            self.iter_to_idx[self.n_individuals] = idx_subj
            self.n_individuals += 1



import warnings
from typing import Optional

import pandas as pd
import torch
from lifelines import WeibullFitter

from leaspy.exceptions import LeaspyInputError
from leaspy.io.data.dataset import Dataset
from leaspy.utils.functional import Exp, MatMul, Sum
from leaspy.utils.typing import DictParams, KwargsType
from leaspy.utils.weighted_tensor import WeightedTensor
from leaspy.variables.distributions import Normal
from leaspy.variables.specs import (
    Hyperparameter,
    LinkedVariable,
    ModelParameter,
    NamedVariables,
    PopulationLatentVariable,
    VariableNameToValueMapping,
)
from leaspy.variables.state import State

from .logistic import LogisticModel
from .obs_models import observation_model_factory

__all__ = ["JointModel"]


class JointModel(LogisticModel):
    """
    Joint model for multiple repeated measures (logistic) and multiple competing events.
    The model implemented is associated to this [publication](https://arxiv.org/abs/2501.08960) on arxiv.

    Parameters
    ----------
    name : :obj:`str`
        The name of the model.

    **kwargs
        Hyperparameters of the model (including `noise_model`)

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        * If `name` is not one of allowed sub-type: 'univariate_linear' or 'univariate_logistic'
        * If hyperparameters are inconsistent
    """

    init_tolerance: float = 0.3

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self._configure_observation_models()
        self._configure_variables_to_track()

    def _configure_variables_to_track(self):
        self.track_variables(["nu", "rho", "nll_attach_y", "nll_attach_event"])
        if self.has_observation_model_with_name("weibull-right-censored-with-sources"):
            self.track_variables(["zeta", "survival_shifts"])
        else:
            self.track_variables(
                [
                    "n_log_nu_mean",
                    "log_rho_mean",
                    "nll_attach_y",
                    "nll_attach_event",
                ]
            )

    def _configure_observation_models(self):
        if (self.dimension == 1) or (self.source_dimension == 0):
            self._configure_univariate_observation_models()
        else:
            self._configure_multivariate_observation_models()

    def _configure_univariate_observation_models(self):
        if self.has_observation_model_with_name("weibull-right-censored-with-sources"):
            raise LeaspyInputError(
                "You cannot use a weibull with sources for an univariate model."
            )
        if not self.has_observation_model_with_name("gaussian-scalar"):
            self.obs_models += (
                observation_model_factory("gaussian-scalar", dimension=1),
            )
        if not self.has_observation_model_with_name("weibull-right-censored"):
            self.obs_models += (
                observation_model_factory(
                    "weibull-right-censored",
                    nu="nu",
                    rho="rho",
                    xi="xi",
                    tau="tau",
                ),
            )

    def _configure_multivariate_observation_models(self):
        if self.has_observation_model_with_name("weibull-right-censored"):
            warnings.warn(
                "You are using a multivariate model with a weibull model without sources"
            )
        elif not self.has_observation_model_with_name(
            "weibull-right-censored-with-sources"
        ):
            self.obs_models += (
                observation_model_factory(
                    "weibull-right-censored-with-sources",
                    nu="nu",
                    rho="rho",
                    zeta="zeta",
                    xi="xi",
                    tau="tau",
                    sources="sources",
                ),
            )

    def get_variables_specs(self) -> NamedVariables:
        """
        Return the specifications of the variables (latent variables, derived variables,
        model 'parameters') that are part of the model.

        Returns
        -------
        NamedVariables :
            The specifications of the model's variables.
        """
        d = super().get_variables_specs()
        d.update(
            # PRIORS
            n_log_nu_mean=ModelParameter.for_pop_mean(
                "n_log_nu",
                shape=(self.nb_events,),
            ),
            n_log_nu_std=Hyperparameter(0.01),
            log_rho_mean=ModelParameter.for_pop_mean(
                "log_rho",
                shape=(self.nb_events,),
            ),
            log_rho_std=Hyperparameter(0.01),
            # LATENT VARS
            n_log_nu=PopulationLatentVariable(
                Normal("n_log_nu_mean", "n_log_nu_std"),
            ),
            log_rho=PopulationLatentVariable(
                Normal("log_rho_mean", "log_rho_std"),
            ),
            # DERIVED VARS
            nu=LinkedVariable(self._exp_neg_n_log_nu),
            rho=LinkedVariable(
                Exp("log_rho"),
            ),
        )
        d.update(
            nll_attach=LinkedVariable(Sum("nll_attach_y", "nll_attach_event")),
            nll_attach_ind=LinkedVariable(
                Sum("nll_attach_y_ind", "nll_attach_event_ind")
            ),
        )
        if self.source_dimension >= 1:
            d.update(
                zeta_mean=ModelParameter.for_pop_mean(
                    "zeta",
                    shape=(self.source_dimension, self.nb_events),
                ),
                zeta_std=Hyperparameter(0.01),
                zeta=PopulationLatentVariable(
                    Normal("zeta_mean", "zeta_std"),
                    sampling_kws={"scale": 0.5},  # cf. GibbsSampler (for retro-compat)
                ),
                survival_shifts=LinkedVariable(MatMul("sources", "zeta")),
            )

        return d

    @staticmethod
    def _exp_neg_n_log_nu(
        *,
        n_log_nu: torch.Tensor,  # TODO: TensorOrWeightedTensor?
    ) -> torch.Tensor:
        """
        Get the scale parameters of the Weibull distribution. This transformation ensures that nu remains positive and simplified update of xi_mean.

        Parameters
        ----------
        n_log_nu : :obj:`torch.Tensor`
            the negative log of the scale parameters of the Weibull distribution (nu)

        Returns
        -------
        torch.Tensor :
            the scale parameters of the Weibull distribution (nu)
        """
        return torch.exp(-1 * n_log_nu)

    @classmethod
    def _center_xi_realizations(cls, state: State) -> None:
        """
        Center the ``xi`` realizations in place.

        .. note::
            This operation does not change the orthonormal basis
            (since the resulting ``v0`` is collinear to the previous one)
            Nor all model computations (only ``v0 * exp(xi_i)`` matters),
            it is only intended for model identifiability / ``xi_i`` regularization
            <!> all operations are performed in "log" space (``v0`` is log'ed)

        Parameters
        ----------
        realizations : :class:`.CollectionRealization`
            The realizations to use for updating the :term:`MCMC` toolbox.
        """
        mean_xi = torch.mean(state["xi"])
        state["xi"] = state["xi"] - mean_xi
        state["log_v0"] = state["log_v0"] + mean_xi
        state["n_log_nu"] = state["n_log_nu"] + mean_xi

    def _load_hyperparameters(self, hyperparameters: KwargsType) -> None:
        """
        Load model's hyperparameters. For joint model it should contain the number of events

        Parameters
        ----------
        hyperparameters : :obj:`dict` [ :obj:`str`, Any ]
            Contains the model's hyperparameters.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If any of the consistency checks fail.
        """
        self.nb_events = hyperparameters.pop("nb_events", 1)
        super()._load_hyperparameters(hyperparameters)

    def to_dict(self, *, with_mixing_matrix: bool = True) -> KwargsType:
        """
        Export model as a dictionary ready for export. Add the number of events compare to the multivariate output

        Parameters
        ----------
        with_mixing_matrix : :obj:`bool`
            If True the mixing matrix will be saved in the dictionary

        Returns
        -------
        KwargsType :
            The model instance serialized as a dictionary.
        """
        dict_params = super().to_dict(with_mixing_matrix=with_mixing_matrix)
        dict_params["nb_events"] = self.nb_events
        return dict_params

    def _validate_compatibility_of_dataset(
        self, dataset: Optional[Dataset] = None
    ) -> None:
        """
        Raise if the given :class:`.Dataset` is not compatible with the current model.

        Parameters
        ----------
        dataset : :class:`.Dataset`, optional

        Raises
        ------
        :exc:`.LeaspyInputError` :
            - If the :class:`.Dataset` has a number of dimensions smaller than 2.
            - If the :class:`.Dataset` does not have the same dimensionality as the model.
            - If the :class:`.Dataset`'s headers do not match the model's.
        """
        super()._validate_compatibility_of_dataset(dataset)
        # Check that there is only one event stored
        if not set(dataset.event_bool.unique().tolist()) == set([False, True]):
            raise LeaspyInputError(
                "You are using a one event model, your event_bool value should only contain 0 and 1, "
                "with at least one censored event and one observed event"
            )

    def _compute_initial_values_for_model_parameters(
        self,
        dataset: Dataset,
    ) -> VariableNameToValueMapping:
        """Compute initial values for model parameters.

        Parameters
        ----------
        dataset : :class:`Dataset`
            Where the individual data are stored

        Returns
        -------
        VariableNameToValueMapping :
            model parameters
        """
        from leaspy.models.utilities import torch_round

        params = super()._compute_initial_values_for_model_parameters(dataset)
        new_parameters = self._estimate_initial_event_parameters(dataset)
        new_rounded_parameters = {
            str(p): torch_round(v.to(torch.float32)) for p, v in new_parameters.items()
        }
        params.update(new_rounded_parameters)
        return params

    def put_individual_parameters(self, state: State, dataset: Dataset):
        """
        Initialise the individual parameters of the state thanks to the dataset.

        Parameters
        ----------
        state : :class:`State`
            where all the variables of the model are stored

        dataset : :class:`Dataset`
            Where the individual data are stored

        Returns
        -------
        None
        """
        df = dataset.to_pandas().reset_index("TIME").groupby("ID").min()
        # Initialise individual parameters if they are not already initialised
        if not state.are_variables_set(("xi", "tau")):
            df_ind = df["TIME"].to_frame(name="tau")
            df_ind["xi"] = 0.0
        else:
            df_ind = pd.DataFrame(
                torch.concat([state["xi"], state["tau"]], axis=1),
                columns=["xi", "tau"],
                index=df.index,
            )
        # Set the right initialisation point for barrier methods
        df_inter = pd.concat(
            [df[dataset.event_time_name] - self.init_tolerance, df_ind["tau"]], axis=1
        )
        df_ind["tau"] = df_inter.min(axis=1)
        if self.source_dimension > 0:
            for i in range(self.source_dimension):
                df_ind[f"sources_{i}"] = 0.0
        with state.auto_fork(None):
            state.put_individual_latent_variables(df=df_ind)

    def _estimate_initial_event_parameters(
        self, dataset: Dataset
    ) -> VariableNameToValueMapping:
        """
        Compute initial values for the event submodel parameters.

        Parameters
        ----------
        dataset : :class:`Dataset`
            Where the individual data are stored

        Returns
        -------
        VariableNameToValueMapping :
            model parameters

        """
        log_rho_mean = [0] * self.nb_events
        n_log_nu_mean = [0] * self.nb_events

        df_ind = dataset.to_pandas().reset_index("TIME").groupby("ID").min()
        approx_tau = torch.tensor(df_ind["TIME"].values) - self.init_tolerance

        for i in range(self.nb_events):
            wbf = WeibullFitter().fit(
                dataset.event_time[:, i] - approx_tau, dataset.event_bool[:, i]
            )
            log_rho_mean[i] = torch.log(torch.tensor(wbf.rho_))
            n_log_nu_mean[i] = -torch.log(torch.tensor(wbf.lambda_))

        event_params = {
            "log_rho_mean": torch.tensor(log_rho_mean),
            "n_log_nu_mean": torch.tensor(n_log_nu_mean),
        }

        if self.source_dimension > 0:
            event_params["zeta_mean"] = torch.zeros(
                self.source_dimension, self.nb_events
            )

        return event_params

    def compute_individual_trajectory(
        self,
        timepoints,
        individual_parameters: DictParams,
        *,
        skip_ips_checks: bool = False,
    ) -> torch.Tensor:
        """
        This method computes the individual trajectory of a patient for given timepoint(s) using his/her individual parameters (random effects).
        For the longitudinal sub-model:
            - Compute longitudinal values
        For the event sub-model:
            - only one event: return the survival rate corrected by the probability of the first time point of the prediction assuming that the patient was alive,
            - more than one event: return the Cumulative Incidence function corrected by the probability of the first time point of the prediction assuming that the patient was alive.
        Nota: model uses its current internal state.

        Parameters
        ----------
        timepoints : scalar or array_like[scalar] (:obj:`list`, :obj:`tuple`, :class:`numpy.ndarray`)
            Contains the age(s) of the subject.
        individual_parameters : :obj:`dict`
            Contains the individual parameters.
            Each individual parameter should be a scalar or array_like.
        skip_ips_checks : :obj:`bool` (default: ``False``)
            Flag to skip consistency/compatibility checks and tensorization
            of ``individual_parameters`` when it was done earlier (speed-up).

        Returns
        -------
        :class:`torch.Tensor`
            Contains the subject's scores computed at the given age(s)
            Shape of tensor is ``(1, n_tpts, n_features)``.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If computation is tried on more than 1 individual.
        :exc:`.LeaspyIndividualParamsInputError`
            if invalid individual parameters.
        """
        self._check_individual_parameters_provided(individual_parameters.keys())
        timepoints, individual_parameters = self._get_tensorized_inputs(
            timepoints, individual_parameters, skip_ips_checks=skip_ips_checks
        )

        # TODO? ability to revert back after **several** assignments?
        # instead of cloning the state for this op?
        local_state = self.state.clone(disable_auto_fork=True)

        self._put_data_timepoints(local_state, timepoints)
        local_state.put(
            "event",
            WeightedTensor(timepoints.T, torch.zeros(timepoints.T.shape).bool()),
        )

        for ip, ip_v in individual_parameters.items():
            local_state[ip] = ip_v
        # reshape survival_event from (len(timepoints)) to (1, len(timepoints), 1) so it is compatible with the
        # model shape
        return torch.cat(
            (
                local_state["model"],
                local_state["predictions_event"].expand(
                    (1, timepoints.shape[1], self.nb_events)
                ),
            ),
            2,
        )



from enum import Enum
from typing import Optional, Union

from .base import BaseModel
from .constant import ConstantModel
from .joint import JointModel
from .linear import LinearModel
from .lme import LMEModel
from .logistic import LogisticModel
from .mixture import LogisticMultivariateMixtureModel
from .shared_speed_logistic import SharedSpeedLogisticModel

__all__ = [
    "ModelName",
    "model_factory",
]


class ModelName(str, Enum):
    """The available models that users can instantiate in Leaspy."""

    JOINT = "joint"
    LOGISTIC = "logistic"
    LINEAR = "linear"
    SHARED_SPEED_LOGISTIC = "shared_speed_logistic"
    LME = "lme"
    CONSTANT = "constant"
    MIXTURE_LOGISTIC = "mixture_logistic"


def model_factory(
    name: Union[str, ModelName], instance_name: Optional[str] = None, **kwargs
) -> BaseModel:
    """
    Return the model object corresponding to ``name`` arg with possible ``kwargs``.

    Parameters
    ----------
    name : :obj:`str` or ModelName
        The name of the model class to be instantiated. Valid options include:
            - ``"joint"``
            - ``"logistic"``
            - ``"linear"``
            - ``"shared_speed_logistic"``
            - ``"lme"``
            - ``"constant"``
            - ``"mixture_logistic"``

    instance_name : :obj:`str`, optional
        A custom name for the model instance. If not provided, the model's name
        will be used as the instance name.

    **kwargs
        Additional keyword arguments corresponding to the model's hyperparameters.
        These must be valid for the specified model, or an error will be raised.

    Returns
    -------
    :class:`.BaseModel`
        A child class object of :class:`.models.BaseModel` class object determined by ``name``.

    Raises
    ------
    ValueError
        If an invalid model name is provided or the model cannot be instantiated
        with the provided arguments.
    """
    name = ModelName(name)
    instance_name = instance_name or name.value
    if name == ModelName.JOINT:
        return JointModel(instance_name, **kwargs)
    if name == ModelName.LOGISTIC:
        return LogisticModel(instance_name, **kwargs)
    if name == ModelName.LINEAR:
        return LinearModel(instance_name, **kwargs)
    if name == ModelName.SHARED_SPEED_LOGISTIC:
        return SharedSpeedLogisticModel(instance_name, **kwargs)
    if name == ModelName.LME:
        return LMEModel(instance_name, **kwargs)
    if name == ModelName.CONSTANT:
        return ConstantModel(instance_name, **kwargs)
    if name == ModelName.MIXTURE_LOGISTIC:
        return LogisticMultivariateMixtureModel(instance_name, **kwargs)



"""Defines the noise model factory."""

from enum import Enum
from typing import Dict, Type, Union

from leaspy.exceptions import LeaspyDataInputError

from .abstract_dataframe_data_reader import AbstractDataframeDataReader
from .covariate_dataframe_data_reader import CovariateDataframeDataReader
from .event_dataframe_data_reader import EventDataframeDataReader
from .joint_dataframe_data_reader import JointDataframeDataReader
from .visit_dataframe_data_reader import VisitDataframeDataReader

__all__ = [
    "DataframeDataReaderNames",
    "DataframeDataReaderFactoryInput",
    "dataframe_data_reader_factory",
]


class DataframeDataReaderNames(Enum):
    """Enumeration defining the possible names for observation models."""

    EVENT = "event"
    VISIT = "visit"
    JOINT = "joint"
    COVARIATE = "covariate"

    @classmethod
    def from_string(cls, reader_name: str):
        """
        Returns the enum member corresponding to the given string.

        Parameters
        ----------
        reader_name : :obj:`str`
            The name of the reader, case-insensitive.

        Returns
        -------
        :class:`~leaspy.io.data.factory.DataframeDataReaderNames`
            The corresponding enum member.

        Raises
        ------
        :exc:`NotImplementedError`
            If the provided `reader_name` does not match any of the enum members and is not implemented.
            Give the valid names in the error message.
        """
        try:
            return cls(reader_name.lower())
        except ValueError:
            raise NotImplementedError(
                f"The requested ObservationModel {reader_name} is not implemented. "
                f"Valid observation model names are: {[elt.value for elt in cls]}."
            )


DataframeDataReaderFactoryInput = Union[
    str, DataframeDataReaderNames, AbstractDataframeDataReader
]

READERS: Dict[DataframeDataReaderNames, Type[AbstractDataframeDataReader]] = {
    DataframeDataReaderNames.EVENT: EventDataframeDataReader,
    DataframeDataReaderNames.VISIT: VisitDataframeDataReader,
    DataframeDataReaderNames.JOINT: JointDataframeDataReader,
    DataframeDataReaderNames.COVARIATE: CovariateDataframeDataReader,
}


def dataframe_data_reader_factory(
    reader: DataframeDataReaderFactoryInput, **kwargs
) -> AbstractDataframeDataReader:
    """
    Factory for observation models.

    Parameters
    ----------
    model : :obj:`str` or :class:`~leaspy.models.obs_models` or :obj:`dict` [ :obj:`str`, ...]
        - If :class:`~leaspy.models.obs_models`, returns the instance.
        - If a string, then returns a new instance of the appropriate class (with optional parameters `kws`).
        - If a dictionary, it must contain the 'name' key and other initialization parameters.
    **kwargs
        Optional parameters for initializing the requested observation model when a string.

    Returns
    -------
    :class:`~leaspy.io.data.abstract_dataframe_data_reader.AbstractDataframeDataReader`
        The desired observation model.

    Raises
    ------
    :exc:`.LeaspyModelInputError` :
        If `model` is not supported.
    """
    if isinstance(reader, AbstractDataframeDataReader):
        return reader
    if isinstance(reader, str):
        reader = DataframeDataReaderNames.from_string(reader)
    if isinstance(reader, DataframeDataReaderNames):
        return READERS[reader](**kwargs)
    raise LeaspyDataInputError(
        "The provided `data_type` should be a valid instance of `DataframeDataReader`, a string "
        f"among {[c.value for c in DataframeDataReaderNames]}."
    )



import torch

from leaspy.exceptions import LeaspyModelInputError

from .abstract_manifold_model_attributes import AbstractManifoldModelAttributes

__all__ = ["LogisticParallelAttributes"]


class LogisticParallelAttributes(AbstractManifoldModelAttributes):
    """
    Attributes of leaspy logistic parallel models.

    Contains the common attributes & methods of the logistic parallel models' attributes.

    Parameters
    ----------
    name : str
    dimension : int
    source_dimension : int

    Attributes
    ----------
    name : str (default 'logistic_parallel')
        Name of the associated leaspy model.
    dimension : int
    source_dimension : int
    has_sources : bool
        Whether model has sources or not (source_dimension >= 1)
    update_possibilities : set[str] (default {'all', 'g', 'deltas', 'betas'} )
        Contains the available parameters to update. Different models have different parameters.
    positions : :class:`torch.Tensor` (scalar) (default None)
        positions = exp(realizations['g']) such that "p0" = 1 / (1 + positions * exp(-deltas))
    deltas : :class:`torch.Tensor` [dimension] (default None)
        deltas = [0, delta_2_realization, ..., delta_n_realization]
    orthonormal_basis : :class:`torch.Tensor` [dimension, dimension - 1] (default None)
    betas : :class:`torch.Tensor` [dimension - 1, source_dimension] (default None)
    mixing_matrix : :class:`torch.Tensor` [dimension, source_dimension] (default None)
        Matrix A such that w_i = A * s_i.

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        if any inconsistent parameters for the model.

    See Also
    --------
    :class:`~leaspy.models.multivariate_parallel_model.MultivariateParallelModel`
    """

    def __init__(self, name, dimension, source_dimension):
        super().__init__(name, dimension, source_dimension)

        if self.dimension < 2:
            raise LeaspyModelInputError(
                f"`LogisticParallelAttributes` with dimension = {self.dimension} (< 2)"
            )

        self.deltas: torch.FloatTensor = (
            None  # deltas = [0, delta_2_realization, ..., delta_n_realization]
        )
        self.update_possibilities = {"all", "g", "deltas"}
        if self.has_sources:
            self.update_possibilities.add("betas")

    def get_attributes(self):
        """
        Returns the following attributes: ``positions``, ``deltas`` & ``mixing_matrix``.

        Returns
        -------
        positions: `torch.Tensor`
        deltas: `torch.Tensor`
        mixing_matrix: `torch.Tensor`
        """
        return self.positions, self.deltas, self.mixing_matrix

    def update(self, names_of_changed_values, values):
        """
        Update model group average parameter(s).

        Parameters
        ----------
        names_of_changed_values : set[str]
            Elements of set must be either:
                * ``all`` (update everything)
                * ``g`` correspond to the attribute :attr:`positions`.
                * ``deltas`` correspond to the attribute :attr:`deltas`.
                * ``betas`` correspond to the linear combination of columns from the orthonormal basis so
                  to derive the :attr:`mixing_matrix`.
        values : dict [str, `torch.Tensor`]
            New values used to update the model's group average parameters

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If `names_of_changed_values` contains unknown parameters.
        """
        self._check_names(names_of_changed_values)

        compute_betas = False
        compute_deltas = False
        compute_positions = False

        if "all" in names_of_changed_values:
            # make all possible updates
            names_of_changed_values = self.update_possibilities

        if "betas" in names_of_changed_values:
            compute_betas = True
        if "deltas" in names_of_changed_values:
            compute_deltas = True
        if "g" in names_of_changed_values:
            compute_positions = True

        if compute_deltas:
            self._compute_deltas(values)
        if compute_positions:
            self._compute_positions(values)

        # only for models with sources beyond this point
        if not self.has_sources:
            return

        if compute_betas:
            self._compute_betas(values)

        # velocities (xi_mean) is a scalar for the logistic parallel model (same velocity for all features - only time-shift the features)
        # so we do not need to compute again the orthonormal basis when updating it (the vector dgamma_t0 stays collinear to previous one)!
        recompute_ortho_basis = compute_positions or compute_deltas

        if recompute_ortho_basis:
            self._compute_orthonormal_basis()
        if recompute_ortho_basis or compute_betas:
            self._compute_mixing_matrix()

    def _compute_positions(self, values):
        """
        Update the attribute ``positions``.

        Parameters
        ----------
        values : dict [str, `torch.Tensor`]
        """
        self.positions = torch.exp(values["g"])

    def _compute_deltas(self, values):
        """
        Update` the attribute ``deltas``.

        Parameters
        ----------
        values : dict [str, `torch.Tensor`]
        """
        self.deltas = torch.cat((torch.tensor([0.0]), values["deltas"]))

    def _compute_gamma_t0_collin_dgamma_t0(self):
        """
        Computes both gamma:
            * value at t0
            * a vector collinear to derivative w.r.t. time at time t0

        Returns
        -------
        2-tuple:
            * gamma_t0 : :class:`torch.Tensor` 1D
            * dgamma_t0 : :class:`torch.Tensor` 1D
        """
        exp_d = torch.exp(-self.deltas)
        denom = 1.0 + self.positions * exp_d
        gamma_t0 = 1.0 / denom

        # we only need a vector which is collinear to dgamma_t0, so we are the laziest possible!
        # we do not multiply by scalars (velocity & position) to get the exact dgamma_t0
        collin_to_dgamma_t0 = exp_d / (denom * denom)
        # collin_to_dgamma_t0 *= self.velocities * self.positions

        return gamma_t0, collin_to_dgamma_t0

    def _compute_orthonormal_basis(self):
        """
        Compute the attribute ``orthonormal_basis`` which is an orthonormal basis, w.r.t the canonical inner product,
        of the sub-space orthogonal, w.r.t the inner product implied by the metric, to the time-derivative of the geodesic at initial time.
        """
        # Compute value and time-derivative of gamma at t0
        gamma_t0, collin_to_dgamma_t0 = self._compute_gamma_t0_collin_dgamma_t0()

        # Compute the diagonal of metric matrix (cf. `_compute_Q`)
        G_metric = (gamma_t0 * (1 - gamma_t0)) ** -2

        # Householder decomposition in non-Euclidean case, updates `orthonormal_basis` in-place
        self._compute_Q(collin_to_dgamma_t0, G_metric)



import re
import time
from pathlib import Path
from typing import Iterable, Optional

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from matplotlib import colormaps
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.lines import Line2D

from leaspy.io.data import Dataset
from leaspy.models import McmcSaemCompatibleModel

from .base import FitAlgorithm


class FitOutputManager:
    """
    Class used by :class:`~leaspy.algo.AbstractAlgo` (and its child classes) to display & save plots and statistics during algorithm execution.

    Parameters
    ----------
    outputs : :class:`~leaspy.algo.OutputsSettings`
        Initialize the `FitOutputManager` class attributes, like the logs paths, the console print periodicity and so forth.

    Attributes
    ----------
    path_output : :obj:`str`
        Path of the folder containing all the outputs
    path_plot : :obj:`str`
        Path of the subfolder of path_output containing the logs plots
    path_plot_convergence_model_parameters : :obj:`str`
        Path of the first plot of the convergence of the model's parameters (in the subfolder path_plot)
    path_plot_patients : :obj:`str`
        Path of the subfolder of path_plot containing the plot of the reconstruction of the patients' longitudinal
        trajectory by the model
    nb_of_patients_to_plot : :obj:`int`
        Number of patients for whom the reconstructions will be plotted.
    path_save_model_parameters_convergence : :obj:`str`
        Path of the subfolder of path_output containing the progression of the model's parameters convergence
    periodicity_plot : :obj:`int` (default 100)
        Set the frequency of the display of the plots
    periodicity_print : :obj:`int`
        Set the frequency of the display of the statistics
    periodicity_save : :obj:`int`
        Set the frequency of the saves of the model's parameters
    periodicity_plot_patients : :obj:`int`
        Set the frequency of the saves of the patients' reconstructions
    plot_sourcewise : :obj:`bool`
        If True, plots will be generated for each source separately.

    """

    def __init__(self, outputs):
        self.periodicity_print = outputs.print_periodicity
        self.periodicity_save = outputs.save_periodicity
        self.periodicity_plot = outputs.plot_periodicity
        self.nb_of_patients_to_plot = outputs.nb_of_patients_to_plot
        self.periodicity_plot_patients = outputs.plot_patient_periodicity
        self.plot_sourcewise = outputs.plot_sourcewise
        if outputs.root_path is not None:
            self.path_output = Path(outputs.root_path)
            self.path_plot = Path(outputs.plot_path)
            self.path_plot_patients = Path(outputs.patients_plot_path)
            self.path_save_model_parameters_convergence = Path(
                outputs.parameter_convergence_path
            )
            self.path_plot_convergence_model_parameters = (
                self.path_plot / "convergence_parameters.pdf"
            )
        self.time = time.time()

    def iteration(
        self, algo: FitAlgorithm, model: McmcSaemCompatibleModel, data: Dataset
    ) -> None:
        """
        Call methods to save state of the running computation, display statistics & plots if the current iteration
        is a multiple of `periodicity_print`, `periodicity_plot` or `periodicity_save`

        Parameters
        ----------
        algo : :class:`~leaspy.algo.fit.FitAlgo`
            A fitting algorithm.

        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            The model used by the computation.

        data : :class:`.Dataset`
            The data used by the computation
        """
        # <!> only `current_iteration` defined for AbstractFitAlgorithm... TODO -> generalize where possible?
        if not hasattr(algo, "current_iteration"):
            # emit a warning?
            return
        iteration = algo.current_iteration

        if self.path_output is None:
            return

        if self.periodicity_print is not None:
            if iteration == 0 or iteration % self.periodicity_print == 0:
                self.print_algo_statistics(algo)
                self.print_model_statistics(model)
                self.print_time()

        if self.periodicity_save is not None:
            if iteration == 0 or iteration % self.periodicity_save == 0:
                self.save_model_parameters_convergence(iteration, model)

        if self.periodicity_plot_patients is not None:
            if iteration == 0 or iteration % self.periodicity_plot_patients == 0:
                self.save_plot_patient_reconstructions(iteration, model, data)

        if self.periodicity_plot is not None:
            if iteration % self.periodicity_plot == 0:
                self.save_plot_convergence_model_parameters(model)

    def print_time(self):
        """Prints the duration since the last periodic point."""
        current_time = time.time()
        print(f"Duration since last print: {current_time - self.time:.3f}s")
        self.time = current_time

    def print_model_statistics(self, model: McmcSaemCompatibleModel):
        """Prints model's statistics.

        Parameters
        ----------
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            The model used by the computation
        """
        print(model)

    def print_algo_statistics(self, algo: FitAlgorithm):
        """Prints algorithm's statistics

        Parameters
        ----------
        algo : :class:`~leaspy.algo.fit.FitAlgo`
            A fitting algorithm.
        """
        print(algo)

    def save_model_parameters_convergence(
        self, iteration: int, model: McmcSaemCompatibleModel
    ) -> None:
        """Saves the current state of the model's parameters

        Parameters
        ----------
        iteration : :obj:`int`
            The current iteration.

        model : :class:`~.models.abstract_model.McmcSaemCompatibleModel`
            The model used by the computation
        """
        model.state.save(
            str(self.path_save_model_parameters_convergence),
            iteration=iteration,
        )

    def save_plot_convergence_model_parameters(self, model: McmcSaemCompatibleModel):
        """Saves figures of the model parameters' convergence in multiple pages of a PDF.

        Parameters
        ----------
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            The model used by the computation
        """
        width = 10
        height_per_row = 3.5

        to_skip = {"betas", "sources", "space_shifts", "xi", "tau", "xi_mean"}
        if model.name == "ordinal":
            to_skip.add("deltas")
        params_with_feature_labels = ["g", "v0"]
        params_with_sources = ["mixing_matrix"]
        params_with_events = []
        if model.name == "joint":
            to_skip.add("survival_shifts")
            params_with_sources.append("zeta")
            params_with_events += ["nu", "rho"]

        params_to_plot = list(model.state.tracked_variables - to_skip)

        files_to_plot = self._get_files_related_to_parameters(params_to_plot)
        # To plot related parameters close to each other, we sort the list
        files_to_plot.sort()

        # If plot sourcewise is true, new sourcewise csv files will be created

        if self.plot_sourcewise and hasattr(model, "source_dimension"):
            files_to_plot = [
                file
                for file in files_to_plot
                if not any(file.name.startswith(param) for param in params_with_sources)
            ]
            files_to_plot.extend(
                self._compute_files_sourcewise(
                    params_with_sources, model.source_dimension
                )
            )

        n_plots = len(files_to_plot)
        with PdfPages(self.path_plot_convergence_model_parameters) as pdf:
            for page in range(0, n_plots, 6):
                # 6 plots per page
                _, ax = plt.subplots(3, 2, figsize=(width, 3 * height_per_row))
                ax = ax.flatten()

                for i, file_path in enumerate(files_to_plot[page : page + 6]):
                    parameter_name = file_path.name.split(".csv")[0]
                    parameter_name, index = self._extract_parameter_name_and_index(
                        parameter_name
                    )
                    df_convergence = pd.read_csv(file_path, index_col=0, header=None)
                    ax[i].plot(df_convergence)

                    ax = self._set_title_for_parameter(
                        ax, i, parameter_name, model, index
                    )

                    ax = self._set_legend_for_parameters(
                        ax,
                        i,
                        parameter_name,
                        params_with_feature_labels,
                        params_with_sources,
                        params_with_events,
                        model,
                    )

                plt.tight_layout()
                pdf.savefig()
                plt.close()

    def save_plot_patient_reconstructions(
        self,
        iteration: int,
        model: McmcSaemCompatibleModel,
        data: Dataset,
    ) -> None:
        """
        Saves figures of real longitudinal values and their reconstructions computed by the model for maximum
        5 patients during each iteration.

        Parameters
        ----------
        iteration : :obj:`int`
            The current iteration
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            The model used by the computation
        data : :class:`~leaspy.io.data.Dataset`
            The dataset used by the computation
        """
        number_of_patient_plot = min(self.nb_of_patients_to_plot, data.n_individuals)
        individual_parameters_dict = {
            variable: model.state.get_tensor_value(variable)
            for variable in model.individual_variables_names
        }

        colors = colormaps["Dark2"](np.linspace(0, 1, number_of_patient_plot + 2))

        fig, ax = plt.subplots(1, 1)
        ax.set_title(f"Feature trajectory for {number_of_patient_plot} patients")
        ax.set_xlabel("Ages")
        ax.set_ylabel("Normalized Feature Value")

        for i in range(number_of_patient_plot):
            times_patient = data.get_times_patient(i).cpu().detach().numpy()
            true_values_patient = data.get_values_patient(i).cpu().detach().numpy()
            ip_patient = {pn: pv[i] for pn, pv in individual_parameters_dict.items()}

            reconstruction_values_patient = (
                model.compute_individual_trajectory(times_patient, ip_patient)
                .squeeze(0)
                .numpy()
            )
            ax.plot(times_patient, reconstruction_values_patient, c=colors[i])
            ax.plot(
                times_patient,
                true_values_patient,
                c=colors[i],
                linestyle="--",
                marker="o",
            )

            last_time_point = times_patient[-1]
            last_reconstruction_value = reconstruction_values_patient.flatten()[-1]
            ax.text(
                last_time_point,
                last_reconstruction_value,
                data.indices[i],
                color=colors[i],
            )

        min_time, max_time = np.percentile(
            data.timepoints[data.timepoints > 0.0].cpu().detach().numpy(),
            [10, 90],
        )
        timepoints_np = np.linspace(min_time, max_time, 100)
        model_values_np = model.compute_mean_traj(
            torch.tensor(np.expand_dims(timepoints_np, 0))
        )

        for feature in range(model.dimension):
            ax.plot(
                timepoints_np,
                model_values_np[0, :, feature],
                c="gray",
                linewidth=3,
                alpha=0.3,
            )

        line_rec = Line2D([0], [0], label="Reconstructions", color="black")
        line_real = Line2D(
            [0], [0], label="Real feature values", color="black", linestyle="--"
        )
        line_avg = Line2D([0], [0], label="Global avg. features", color="gray")

        handles, labels = ax.get_legend_handles_labels()
        handles.extend([line_rec, line_real, line_avg])

        ax.legend(handles=handles)
        path_iteration = self.path_plot_patients / f"plot_patients_{iteration}.pdf"

        plt.savefig(path_iteration)
        plt.close()

    def _get_files_related_to_parameters(self, parameters: Iterable[str]) -> list[Path]:
        """Retrieve the list of file paths related to the given parameters.

        Parameters
        ----------
        parameters : Iterable[:obj:`str`]
            A list of parameter names.

        Returns
        -------
        list[Path]
            A list of file paths that match the parameter names.
        """
        return [
            f
            for f in self.path_save_model_parameters_convergence.iterdir()
            if any(f.name.startswith(param) for param in parameters)
        ]

    def _extract_parameter_name_and_index(
        self, parameter_name: str
    ) -> tuple[Optional[str], Optional[int]]:
        """Extract the parameter name and its corresponding index (if applicable) from the given parameter name.

         Parameters
         ----------
         parameter_name : :obj:`str`
             The name of the parameter to extract information from.

         Returns
         -------
         tuple[Optional[:obj:`str`], Optional[:obj:`int`]]
             A tuple where the first element is the parameter name and the second is its index (if applicable).

         Examples
         --------
         >>> _extract_parameter_name_and_index("mixing_matrix_10")
        ('mixing_matrix', 10)
         >>> _extract_parameter_name_and_index("mixing_matrix_1")
         ('mixing_matrix', 1)
         >>> _extract_parameter_name_and_index("mixing_matrix_")
         ('mixing_matrix_', None)
         >>> _extract_parameter_name_and_index("mixing_matrix_10_foo")
        ('mixing_matrix_10_foo', None)
        >>> _extract_parameter_name_and_index("v0")
        ('v0', None)
        """
        if parameter_name == "v0":
            return parameter_name, None
        match = re.search(r"^(.*?)(\d+)$", parameter_name)
        if match:
            return match.group(1).strip("_"), int(match.group(2))
        else:
            return parameter_name, None

    def _set_title_for_parameter(
        self,
        ax: plt.Axes,
        i: int,
        parameter_name: str,
        model: McmcSaemCompatibleModel,
        index: int,
    ) -> plt.Axes:
        """Set the title for the plot based on the parameter name and the model's features or other information.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes object of the plot.
        i : :obj:`int`
            The index of the current plot.
        parameter_name : :obj:`str`
            The name of the parameter being plotted.
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            The model containing features and event information.
        index : :obj:`int`
            The index of the specific feature or event, or None if not applicable.

        Returns
        -------
        matplotlib.axes.Axes
            The axes object with the title set.
        """
        if parameter_name == "mixing_matrix":
            ax[i].set_title(parameter_name + " " + model.features[index])
        elif parameter_name == "zeta":
            ax[i].set_title(parameter_name + " " + "event" + " " + str(index + 1))
        elif parameter_name.startswith("sourcewise"):
            ax[i].set_title(
                parameter_name.replace("sourcewise_", "")
                + " "
                + "source"
                + " "
                + str(index)
            )
        else:
            ax[i].set_title(parameter_name)
        return ax

    def _set_legend_for_parameters(
        self,
        ax: plt.Axes,
        i: int,
        parameter_name: str,
        params_with_feature_labels: Iterable[str],
        params_with_sources: Iterable[str],
        params_with_events: Iterable[str],
        model: McmcSaemCompatibleModel,
    ) -> plt.Axes:
        """Set the legend for the plot based on the parameter name and the model's features, sources, or events.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes object of the plot.
        i : :obj:`int`
            The index of the current plot.
        parameter_name : :obj:`str`
            The name of the parameter being plotted.
        params_with_feature_labels : list[:obj:`str`]
            A list of parameters associated with feature labels.
        params_with_sources : list[:obj:`str`]
            A list of parameters associated with sources.
        params_with_events : list[:obj:`str`]
            A list of parameters associated with events.
        model : :class:`~leaspy.models.McmcSaemCompatibleModel`
            The model containing the necessary information for legends.

        Returns
        -------
        matplotlib.axes.Axes
            The axes object with the legend set.
        """
        if parameter_name in params_with_feature_labels:
            ax[i].legend(model.features, loc="best")
        if hasattr(model, "source_dimension") and parameter_name in params_with_sources:
            sources = [
                "Source" + " " + str(i + 1) for i in range(model.source_dimension)
            ]
            ax[i].legend(sources, loc="best")
        if hasattr(model, "nb_events"):
            if parameter_name in params_with_events:
                events = ["Event" + " " + str(i + 1) for i in range(model.nb_events)]
                ax[i].legend(events, loc="best")
            if parameter_name.startswith("sourcewise"):
                if "mixing_matrix" in parameter_name:
                    ax[i].legend(model.features, loc="best")
                if "zeta" in parameter_name:
                    events = [
                        "Event" + " " + str(i + 1) for i in range(model.nb_events)
                    ]
                    ax[i].legend(events, loc="best")
        return ax

    def _compute_files_sourcewise(
        self, params_with_sources: Iterable[str], num_sources: int
    ) -> list[Path]:
        """This function processes parameters that have sources and generates new csv files for each source.

        Parameters
        ----------
        params_with_sources : list[:obj:`str`]
            A list of parameters that have associated source data.
        num_sources : :obj:`int`
            The number of sources in model

        Returns
        -------
        list[Path]
            The list of newly created sourcewise files.
        """
        new_files = []
        for param_name in params_with_sources:
            related_files = self._get_files_related_to_parameters([param_name])
            if not related_files:
                continue
            related_files.sort()

            for source_idx in range(num_sources):
                new_files.append(
                    self._concatenate_and_save_sourcewise_parameters(
                        param_name, related_files, source_idx
                    )
                )
        return new_files

    def _concatenate_and_save_sourcewise_parameters(
        self, parameter_name: str, files: Iterable[Path], source_idx: int
    ) -> Path:
        """Concatenates data for the specific source and saves it as a csv file.

        Parameters
        ----------
        parameter_name : :obj:`str`
              A sourcewise parameter
        files : Iterable[Path]
              A list of file paths containing sourcewise parameter
        source_idx : :obj:`int`
              The index of the source
        Returns
        -------
        Path
            The path to the saved csv file containing data for the specified source
        """
        df = self._concatenate_parameters(files, source_idx)
        file_name = f"sourcewise_{parameter_name}_{source_idx + 1}.csv"
        file_path = self.path_save_model_parameters_convergence / file_name
        df.to_csv(file_path, header=False)
        return file_path

    def _concatenate_parameters(
        self, files: Iterable[Path], source_idx: int
    ) -> pd.DataFrame:
        """Extracts and concatenates a specific column from multiple csv files into a single DataFrame

        Parameters
        ----------
        files : Iterable[Path]
            A list of file paths
        source_idx : :obj: `int`
            The index of the source

        Returns
        -------
        :obj: `pd.DataFrame`
            A DataFrame with each column representing the selected source's data from multiple files
        """
        return pd.concat(
            [
                pd.read_csv(file_path, index_col=0, header=None).iloc[:, source_idx]
                for file_path in files
            ],
            axis=1,
            join="inner",
        )



import warnings
from typing import Optional

import pandas as pd

from leaspy.exceptions import LeaspyDataInputError
from leaspy.utils.typing import FeatureType

from .abstract_dataframe_data_reader import AbstractDataframeDataReader
from .event_dataframe_data_reader import EventDataframeDataReader
from .individual_data import IndividualData
from .visit_dataframe_data_reader import VisitDataframeDataReader

__all__ = ["JointDataframeDataReader"]


class JointDataframeDataReader(AbstractDataframeDataReader):
    """
    Methods to convert :class:`pandas.DataFrame` to `Leaspy`-compliant data containers for event data and longitudinal data.

    Parameters
    ----------
    event_time_name: str
        Name of the columns in dataframe that contains the time of event
    event_bool_name: str
        Name of the columns in dataframe that contains if the event is censored of not

    Raises
    ------
    :exc:`.LeaspyDataInputError`
    """

    tol_diff = 0.001

    def __init__(
        self,
        *,
        event_time_name: str = "EVENT_TIME",
        event_bool_name: str = "EVENT_BOOL",
        nb_events: Optional[int] = None,
    ):
        super().__init__()

        self.visit_reader = VisitDataframeDataReader()
        self.event_reader = EventDataframeDataReader(
            event_time_name=event_time_name,
            event_bool_name=event_bool_name,
            nb_events=nb_events,
        )

    ######################################################
    #               JOINT METHODS
    ######################################################
    @property
    def event_time_name(self) -> str:
        """Name of the event time column in dataset"""
        return self.event_reader.event_time_name

    @property
    def event_bool_name(self) -> str:
        """Name of the event bool column in dataset"""
        return self.event_reader.event_bool_name

    @property
    def dimension(self) -> Optional[int]:
        """Number of longitudinal outcomes in dataset."""
        return self.visit_reader.dimension

    @property
    def long_outcome_names(self) -> list[FeatureType]:
        """Name of the longitudinal outcomes in dataset"""
        return self.visit_reader.long_outcome_names

    @property
    def n_visits(self) -> int:
        """Number of visit in the dataset"""
        return self.visit_reader.n_visits

    def _check_headers(self, columns: list[str]) -> None:
        """
        Check mendatory dataframe headers

        Parameters
        ----------
        columns: List[str]
            Names of the columns headers of the dataframe that contains patients information
        """
        self.visit_reader._check_headers(columns)
        self.event_reader._check_headers(columns)

    def _set_index(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Set the index suited for the type of information contained in the dataframe

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information

        Returns
        -------
        df: pd.DataFrame
            Dataframe with the right index
        """

        return self.visit_reader._set_index(df)

    def _clean_dataframe(
        self, df: pd.DataFrame, *, drop_full_nan: bool, warn_empty_column: bool
    ) -> pd.DataFrame:
        """
        Clean the dataframe that contains patient information which are repeated measures and events

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information.

        drop_full_nan: bool
            If set to True, raw full of nan are dropped.

        warn_empty_column: bool
            If set to True, a warning is raise for columns full of nan.


        Returns
        -------
        df: pd.DataFrame
            Dataframe with clean information.
        """

        # Check visits
        df_visit = self.visit_reader._clean_dataframe(
            df.drop([self.event_time_name, self.event_reader.event_bool_name], axis=1),
            drop_full_nan=drop_full_nan,
            warn_empty_column=warn_empty_column,
        )

        # Check events
        df_event = self.event_reader._clean_dataframe(
            df.reset_index()
            .drop(self.long_outcome_names + ["TIME"], axis=1)
            .set_index("ID"),
            drop_full_nan=drop_full_nan,
            warn_empty_column=warn_empty_column,
        )

        # [SPECIFIC] prepare_clean_output
        if (
            not df_event.groupby("ID")
            .first()
            .index.equals(df_visit.groupby("ID").first().index)
        ):
            raise LeaspyDataInputError(
                "All patients must have at least one visit and one event"
            )

        df = df_visit.join(df_event)

        # Additional crossed check
        df_test = df.reset_index().groupby("ID").max()
        if not (
            df_test[self.event_time_name] - df_test["TIME"] >= -self.tol_diff
        ).all():
            df_before = df_test[
                ~(df_test[self.event_time_name] - df_test["TIME"] >= -self.tol_diff)
            ]
            if df_before[self.event_bool_name].sum() == 0:
                warnings.warn(
                    "You have event censored before the last available visits, you should be in a prediction set-up"
                )
            else:
                raise LeaspyDataInputError(
                    f"Event should happen after or at the last visit "
                    f"for {df_before.index.tolist()} patients"
                )

        return df

    def _load_individuals_data(
        self, subj: IndividualData, df_subj: pd.DataFrame
    ) -> None:
        """
        Convert information stored in a dataframe to information stored into IndividualData

        Parameters
        ----------
        subj: IndividualData
            One patient with her/his information, potentially empty

        df_subj: pd.DataFrame
            One patient with her/his information
        """
        self.visit_reader._load_individuals_data(subj, df_subj)
        self.event_reader._load_individuals_data(subj, df_subj)



import abc
from collections.abc import Sequence
from random import shuffle
from typing import ClassVar, Union

import torch
from numpy import ndindex

from leaspy.exceptions import LeaspyInputError
from leaspy.variables.state import State

from .base import AbstractIndividualSampler, AbstractPopulationSampler

__all__ = [
    "IteratorIndicesType",
    "GibbsSamplerMixin",
    "AbstractPopulationGibbsSampler",
    "PopulationGibbsSampler",
    "PopulationFastGibbsSampler",
    "PopulationMetropolisHastingsSampler",
    "IndividualGibbsSampler",
]

IteratorIndicesType = list[tuple[int, ...]]


class GibbsSamplerMixin:
    """
    Mixin class for Gibbs samplers (individual and population).

    This class contains common logic for all Gibbs samplers.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    scale : :obj:`float` > 0 or :class:`torch.FloatTensor` > 0
        An approximate scale for the variable.
        It will be used to scale the initial adaptive std-dev used in sampler.
        An extra factor will be applied on top of this scale :
            * 1% for population parameters (:attr:`.AbstractPopulationGibbsSampler.STD_SCALE_FACTOR`)
            * 50% for individual parameters (:attr:`.IndividualGibbsSampler.STD_SCALE_FACTOR`)
        Note: If a :class:`torch.Tensor` is passed, its shape must be compatible with the shape of the variable.
    random_order_dimension : :obj:`bool` (default True)
        Whether to randomize the order of dimensions during the sampling loop.
        This is applied only to population variables as group sampling is used for individual variables.
        `Article <https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html>`_
        gives a reason on why we should activate this flag.
    mean_acceptation_rate_target_bounds : :obj:`tuple` [lower_bound: :obj:`float`, upper_bound: :obj:`float`] with 0 < lower_bound < upper_bound < 1
        Bounds on the mean acceptation rate.
        If the rate is outside this range, the adaptive std-dev will be updated
        to maintain the target acceptance rate. (e.g: ~30%).
    adaptive_std_factor : :obj:`float` in (0, 1)
        The factor by which the sampler's std-dev is adapted:

        * If acceptance rate is too low → decrease by ``1 - factor``
        * If too high → increase by ``1 + factor``
    **base_sampler_kws
        Additional keyword arguments passed to :class:`~leaspy.samplers.AbstractSampler` init method.
        For example, you may pass `acceptation_history_length`.

    Attributes
    ----------
    In addition to the attributes present in AbstractSampler :

    std : :class:`torch.Tensor`
        Adaptive std-dev of variable

    Raises
    ------
    :exc:`.LeaspyInputError`
    """

    STD_SCALE_FACTOR: ClassVar[float]

    def __init__(
        self,
        name: str,
        shape: tuple,
        *,
        scale: Union[float, torch.FloatTensor],
        random_order_dimension: bool = True,
        mean_acceptation_rate_target_bounds: tuple[float, float] = (0.2, 0.4),
        adaptive_std_factor: float = 0.1,
        **base_sampler_kws,
    ):
        super().__init__(name, shape, **base_sampler_kws)
        self.scale = self.validate_scale(scale)
        self.std = (
            self.STD_SCALE_FACTOR * self.scale * torch.ones(self.shape_adapted_std)
        )
        # Internal counter to trigger adaptation of std based on mean acceptation rate
        self._counter: int = 0
        # Parameters of the sampler
        self._random_order_dimension = random_order_dimension
        self._set_acceptation_bounds(mean_acceptation_rate_target_bounds)
        self._set_adaptive_std_factor(adaptive_std_factor)

    @property
    @abc.abstractmethod
    def shape_adapted_std(self) -> tuple[int, ...]:
        """Shape of adaptative variance."""

    @property
    def shape_acceptation(self) -> tuple[int, ...]:
        """Shape of adaptative variance."""
        return self.shape_adapted_std

    def _set_acceptation_bounds(
        self, mean_acceptation_rate_target_bounds: tuple[float, float]
    ) -> None:
        if not (
            isinstance(mean_acceptation_rate_target_bounds, Sequence)
            and len(mean_acceptation_rate_target_bounds) == 2
            and 0
            < mean_acceptation_rate_target_bounds[0]
            < mean_acceptation_rate_target_bounds[1]
            < 1
        ):
            raise LeaspyInputError(
                "`mean_acceptation_rate_target_bounds` should be a tuple (lower_bound, upper_bound) "
                f"with 0 < lower_bound < upper_bound < 1, not '{mean_acceptation_rate_target_bounds}'"
            )
        (
            self._mean_acceptation_lower_bound_before_adaptation,
            self._mean_acceptation_upper_bound_before_adaptation,
        ) = mean_acceptation_rate_target_bounds

    def _set_adaptive_std_factor(self, adaptive_std_factor: float) -> None:
        if not (0 < adaptive_std_factor < 1):
            raise LeaspyInputError(
                f"`adaptive_std_factor` should be a float in ]0, 1[, not '{adaptive_std_factor}'"
            )
        self._adaptive_std_factor = adaptive_std_factor

    def validate_scale(self, scale: Union[float, torch.Tensor]) -> torch.Tensor:
        """
        Validate the user provided scale if it is a :obj:`float` or :class:`torch.Tensor`.

        The scale must be positive. If the input is multidimensional, all components must be positive.

        Parameters
        ----------
        scale : :obj:`float` or :class:`torch.Tensor`
            The scale to be validated.

        Returns
        -------
        :class:`torch.Tensor` :
            Valid scale.

        Raises
        ------
        :exc:`.LeaspyInputError`
            If the scale contains any non-positive value.
        """
        if not isinstance(scale, torch.Tensor):
            scale = torch.tensor(scale)
        scale = scale.float()
        if (scale <= 0).any():
            raise LeaspyInputError(
                f"Scale of variable '{self.name}' should be positive, not `{scale}`."
            )
        return scale

    def __str__(self):
        mean_std = self.std[self._meaningful_indices].mean()
        mean_acceptation_rate = self.acceptation_history[
            (slice(None),) + self._meaningful_indices
        ].mean()
        return f"{self.name} rate : {mean_acceptation_rate.item():.1%}, std: {mean_std.item():.1e}"

    @property
    def _meaningful_indices(self) -> tuple[torch.Tensor, ...]:
        """
        Return the subset of indices that are relevant for both adapted-variance
        and acceptations tensors.

        By default, this is an empty :obj:`tuple` which corresponds to no masking at all
        of adapted-variances nor acceptations tensors.

        Returns
        -------
        :obj:`tuple` of :obj:`int` :
            The meaningful indices.
        """
        return ()

    def _update_std(self):
        """
        Update standard deviation of sampler according to current frequency of acceptation.

        Adaptive std is known to improve sampling performances.

        For default parameters:
            - `std-dev` is increased if frequency of acceptation is > 40%
            - `std-dev` is decreased if frequency of acceptation is < 20%
              (to stay close to 30%).
        """
        self._counter += 1

        if self._counter % self.acceptation_history_length == 0:
            mean_acceptation = self.acceptation_history.mean(dim=0)

            # nota: for masked elements in full Gibbs, std will always remain = 0
            idx_toolow = (
                mean_acceptation < self._mean_acceptation_lower_bound_before_adaptation
            )
            idx_toohigh = (
                mean_acceptation > self._mean_acceptation_upper_bound_before_adaptation
            )

            self.std[idx_toolow] *= 1 - self._adaptive_std_factor
            self.std[idx_toohigh] *= 1 + self._adaptive_std_factor


class AbstractPopulationGibbsSampler(GibbsSamplerMixin, AbstractPopulationSampler):
    """
    Abstract class for all Gibbs samplers for population variables.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    scale : :obj:`float` > 0 or :class:`torch.FloatTensor` > 0
        An approximate scale for the variable.
        It will be used to scale the initial adaptive std-dev used in sampler.
        An extra factor will be applied on top of this scale :
            * 1% for population parameters (:attr:`.AbstractPopulationGibbsSampler.STD_SCALE_FACTOR`)
            * 50% for individual parameters (:attr:`.IndividualGibbsSampler.STD_SCALE_FACTOR`)
        Note: If a :class:`torch.Tensor` is passed, its shape must be compatible with the shape of the variable.
    random_order_dimension : :obj:`bool` (default True)
        Whether to randomize the order of dimensions during the sampling loop.
        This is applied only to population variables as group sampling is used for individual variables.
        `Article <https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html>`_
        gives a reason on why we should activate this flag.
    mean_acceptation_rate_target_bounds : :obj:`tuple` [lower_bound: :obj:`float`, upper_bound: :obj:`float`] with 0 < lower_bound < upper_bound < 1
        Bounds on the mean acceptation rate.
        If the rate is outside this range, the adaptive std-dev will be updated
        to maintain the target acceptance rate. (e.g: ~30%).
    adaptive_std_factor : :obj:`float` in (0, 1)
        The factor by which the sampler's std-dev is adapted:

        * If acceptance rate is too low → decrease by ``1 - factor``
        * If too high → increase by ``1 + factor``
    **base_sampler_kws
        Additional keyword arguments passed to :class:`~leaspy.samplers.AbstractSampler` init method.
        For example, you may pass `acceptation_history_length`.

    """

    STD_SCALE_FACTOR = 0.01

    def validate_scale(self, scale: Union[float, torch.Tensor]) -> torch.Tensor:
        """
        Validate the user provided scale in :obj:`float` or :class:`torch.Tensor`.

        If necessary, the scale is cast to a :class:`torch.Tensor`.

        Parameters
        ----------
        scale : :obj:`float` or :class:`torch.Tensor`
            The scale to be validated.

        Returns
        -------
        :class:`torch.Tensor` :
            Valid scale.
        """
        scale = super().validate_scale(scale)
        if scale.ndim > len(self.shape_adapted_std):
            # we take the mean of grouped dimension in this case
            scale_squeezed_dims = tuple(range(len(self.shape_adapted_std), scale.ndim))
            scale = scale.mean(dim=scale_squeezed_dims)
        return scale

    def sample(
        self,
        state: State,
        *,
        temperature_inv: float,
    ) -> None:
        """
        Apply a Metropolis-Hastings (MH) sampling step for each dimension of the population variable.

        For each dimension (1D or 2D):
         - Compute the current attachment and regularity terms.
         - Propose a new value for the dimension.
         - Recompute the attachment and regularity based on the new value.
         - Apply a Metropolis-Hastings step to accept or reject the proposal.

        Parameters
        ----------
        state : :class:`.State`
            Object containing values for all model variables, including latent variables
        temperature_inv : :obj:`float` > 0
            The temperature to use.
        """
        accepted_array = torch.zeros_like(self.std)

        def compute_attachment_regularity():
            # Mask for regularity is handled directly by `WeightedTensor` logic
            return state["nll_attach"], state[f"nll_regul_{self.name}"]

        for idx in self._get_shuffled_iterator_indices():
            previous_attachment, previous_regularity = compute_attachment_regularity()
            # with state.auto_fork():  # not needed since state already have auto_fork on
            state.put(
                self.name,
                self._proposed_change_idx(idx),
                indices=idx,
                accumulate=True,  # out-of-place addition
            )

            # Update (and caching) of derived model attributes (orthonormal basis, ...) is done in state
            new_attachment, new_regularity = compute_attachment_regularity()
            alpha = torch.exp(
                -1
                * (
                    (new_regularity - previous_regularity) * temperature_inv
                    + (new_attachment - previous_attachment)
                )
            )
            accepted = self._metropolis_step(alpha)
            accepted_array[idx] = accepted

            if not accepted:
                state.revert()

        self._update_acceptation_rate(accepted_array)
        self._update_std()

    def _get_shuffled_iterator_indices(self) -> IteratorIndicesType:
        indices = self._get_iterator_indices()
        if self._random_order_dimension:
            shuffle(indices)  # shuffle in-place!
        return indices

    def _get_iterator_indices(self) -> IteratorIndicesType:
        """
        When a mask is set on the random variable and when we sample without any grouping
        (i.e. regular Gibbs sampler - i.e. shape of `std` is same as shape of `mask`)
        then we'll only loop on coordinates not masked (we skip the un-needed indices)

        Depending on sampler type we will loop on all coordinates or (partially) group them:
        Example for variable of shape (2, 3)
            * for 'Gibbs' sampler: [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]
            * for 'FastGibbs' sampler: [(0,), (1,)]
            * for 'Metropolis-Hastings' sampler: [()]

        Returns
        -------
        IteratorIndicesType :
            The indices for the iterator.
        """
        return list(ndindex(self.shape_adapted_std))

    @property
    def _should_mask_changes(self) -> bool:
        return self.mask is not None

    def _proposed_change_idx(self, idx: tuple[int, ...]) -> torch.Tensor:
        """
        The previous version with `_proposal` was not incorrect but computationally inefficient:
        because we were sampling on the full shape of `std` whereas we only needed `std[idx]` (smaller)

        We don't directly mask the `new_val_idx` since it may be infinite, producing nans
        when trying to multiply them by 0.

        Returns
        -------
        :class:`torch.Tensor` :
            The proposed change for the given index.
        """
        shape_idx = self.shape[slice(len(idx), self.ndim)]
        change_idx = self.std[idx] * torch.randn(shape_idx)
        if self._should_mask_changes:
            change_idx = change_idx * self.mask[idx].float()
        return change_idx


class PopulationGibbsSampler(AbstractPopulationGibbsSampler):
    """
    Gibbs sampler for population variables.

    The sampling is done iteratively for all coordinate values.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    scale : :obj:`float` > 0 or :class:`torch.FloatTensor` > 0
        An approximate scale for the variable.
        It will be used to scale the initial adaptive std-dev used in sampler.
        An extra factor will be applied on top of this scale :
            * 1% for population parameters (:attr:`.AbstractPopulationGibbsSampler.STD_SCALE_FACTOR`)
            * 50% for individual parameters (:attr:`.IndividualGibbsSampler.STD_SCALE_FACTOR`)
        Note: If a :class:`torch.Tensor` is passed, its shape must be compatible with the shape of the variable.
    random_order_dimension : :obj:`bool` (default True)
        Whether to randomize the order of dimensions during the sampling loop.
        This is applied only to population variables as group sampling is used for individual variables.
        `Article <https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html>`_
        gives a reason on why we should activate this flag.
    mean_acceptation_rate_target_bounds : :obj:`tuple` [lower_bound: :obj:`float`, upper_bound: :obj:`float`] with 0 < lower_bound < upper_bound < 1
        Bounds on the mean acceptation rate.
        If the rate is outside this range, the adaptive std-dev will be updated
        to maintain the target acceptance rate. (e.g: ~30%).
    adaptive_std_factor : :obj:`float` in (0, 1)
        The factor by which the sampler's std-dev is adapted:

        * If acceptance rate is too low → decrease by ``1 - factor``
        * If too high → increase by ``1 + factor``
    **base_sampler_kws
        Additional keyword arguments passed to :class:`~leaspy.samplers.AbstractSampler` init method.
        For example, you may pass `acceptation_history_length`.
    """

    def __init__(
        self,
        name: str,
        shape: tuple,
        *,
        scale: Union[float, torch.FloatTensor],
        random_order_dimension: bool = True,
        mean_acceptation_rate_target_bounds: tuple[float, float] = (0.2, 0.4),
        adaptive_std_factor: float = 0.1,
        **base_sampler_kws,
    ):
        super().__init__(
            name,
            shape,
            scale=scale,
            random_order_dimension=random_order_dimension,
            mean_acceptation_rate_target_bounds=mean_acceptation_rate_target_bounds,
            adaptive_std_factor=adaptive_std_factor,
            **base_sampler_kws,
        )
        # adapted-variance of mask elements are meaningless (not taken into account for aggregated stats)
        if self.mask is not None:
            self.std[~self.mask] = 0

    @property
    def _should_mask_changes(self) -> bool:
        """
        Nota: for full Gibbs, strictly speaking we never need to applying any masking on proposed changes
        (since we already forced std=0 on masked elements if any; it would not hurt but would be slightly less efficient).
        """
        return False

    @property
    def shape_adapted_std(self) -> tuple:
        return self.shape

    @property
    def _meaningful_indices(self) -> tuple:
        if self.mask is not None:
            return (self.mask,)
        return ()

    def _get_iterator_indices(self) -> IteratorIndicesType:
        """
        Example for variable of shape (2,3) with mask = [[1,1,0],[1,1,1]]
            --> iterator_indices = [(0,0), (0, 1), (1, 0), (1, 1), (1, 2)]

        Returns
        -------
        IteratorIndicesType :
            The indices for the iterator.
        """
        if self.mask is not None:
            return list(map(tuple, self.mask.nonzero(as_tuple=False).tolist()))
        return super()._get_iterator_indices()


class PopulationFastGibbsSampler(AbstractPopulationGibbsSampler):
    """
    Fast Gibbs sampler for population variables.

    .. note::
        The sampling batches along the dimensions except the first one.
        This accelerates sampling process for 2D parameters.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    scale : :obj:`float` > 0 or :class:`torch.FloatTensor` > 0
        An approximate scale for the variable.
        It will be used to scale the initial adaptive std-dev used in sampler.
        An extra factor will be applied on top of this scale :
            * 1% for population parameters (:attr:`.AbstractPopulationGibbsSampler.STD_SCALE_FACTOR`)
            * 50% for individual parameters (:attr:`.IndividualGibbsSampler.STD_SCALE_FACTOR`)
        Note: If a :class:`torch.Tensor` is passed, its shape must be compatible with the shape of the variable.
    random_order_dimension : :obj:`bool` (default True)
        Whether to randomize the order of dimensions during the sampling loop.
        This is applied only to population variables as group sampling is used for individual variables.
        `Article <https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html>`_
        gives a reason on why we should activate this flag.
    mean_acceptation_rate_target_bounds : :obj:`tuple` [lower_bound: :obj:`float`, upper_bound: :obj:`float`] with 0 < lower_bound < upper_bound < 1
        Bounds on the mean acceptation rate.
        If the rate is outside this range, the adaptive std-dev will be updated
        to maintain the target acceptance rate. (e.g: ~30%).
    adaptive_std_factor : :obj:`float` in (0, 1)
        The factor by which the sampler's std-dev is adapted:

        * If acceptance rate is too low → decrease by ``1 - factor``
        * If too high → increase by ``1 + factor``
    **base_sampler_kws
        Additional keyword arguments passed to :class:`~leaspy.samplers.AbstractSampler` init method.
        For example, you may pass `acceptation_history_length`.
    """

    @property
    def shape_adapted_std(self) -> tuple:
        return (self.shape[0],)


class PopulationMetropolisHastingsSampler(AbstractPopulationGibbsSampler):
    """
    Metropolis-Hastings sampler for population variables.

    .. note::
        The sampling is implemented for all values at once.
        This accelerates sampling process but usually requires more iterations.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    scale : :obj:`float` > 0 or :class:`torch.FloatTensor` > 0
        An approximate scale for the variable.
        It will be used to scale the initial adaptive std-dev used in sampler.
        An extra factor will be applied on top of this scale :
            * 1% for population parameters (:attr:`.AbstractPopulationGibbsSampler.STD_SCALE_FACTOR`)
            * 50% for individual parameters (:attr:`.IndividualGibbsSampler.STD_SCALE_FACTOR`)
        Note: If a :class:`torch.Tensor` is passed, its shape must be compatible with the shape of the variable.
    random_order_dimension : :obj:`bool` (default True)
        Whether to randomize the order of dimensions during the sampling loop.
        This is applied only to population variables as group sampling is used for individual variables.
        `Article <https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html>`_
        gives a reason on why we should activate this flag.
    mean_acceptation_rate_target_bounds : :obj:`tuple` [lower_bound: :obj:`float`, upper_bound: :obj:`float`] with 0 < lower_bound < upper_bound < 1
        Bounds on the mean acceptation rate.
        If the rate is outside this range, the adaptive std-dev will be updated
        to maintain the target acceptance rate. (e.g: ~30%).
    adaptive_std_factor : :obj:`float` in (0, 1)
        The factor by which the sampler's std-dev is adapted:

        * If acceptance rate is too low → decrease by ``1 - factor``
        * If too high → increase by ``1 + factor``
    **base_sampler_kws
        Additional keyword arguments passed to :class:`~leaspy.samplers.AbstractSampler` init method.
        For example, you may pass `acceptation_history_length`.
    """

    @property
    def shape_adapted_std(self) -> tuple:
        return ()


class IndividualGibbsSampler(GibbsSamplerMixin, AbstractIndividualSampler):
    """
    Gibbs sampler for individual variables.

    Individual variables are handled using a grouped Gibbs sampler.
    Currently, this is the only sampler implemented for individual variables.

    Parameters
    ----------
    name : :obj:`str`
        The name of the random variable to sample.
    shape : :obj:`tuple` of :obj:`int`
        The shape of the random variable to sample.
    scale : :obj:`float` > 0 or :class:`torch.FloatTensor` > 0
        An approximate scale for the variable.
        It will be used to scale the initial adaptive std-dev used in sampler.
        An extra factor will be applied on top of this scale :
            * 1% for population parameters (:attr:`.AbstractPopulationGibbsSampler.STD_SCALE_FACTOR`)
            * 50% for individual parameters (:attr:`.IndividualGibbsSampler.STD_SCALE_FACTOR`)
        Note: If a :class:`torch.Tensor` is passed, its shape must be compatible with the shape of the variable.
    random_order_dimension : :obj:`bool` (default True)
        Whether to randomize the order of dimensions during the sampling loop.
        This is applied only to population variables as group sampling is used for individual variables.
        `Article <https://proceedings.neurips.cc/paper/2016/hash/e4da3b7fbbce2345d7772b0674a318d5-Abstract.html>`_
        gives a reason on why we should activate this flag.
    mean_acceptation_rate_target_bounds : :obj:`tuple` [lower_bound: :obj:`float`, upper_bound: :obj:`float`] with 0 < lower_bound < upper_bound < 1
        Bounds on the mean acceptation rate.
        If the rate is outside this range, the adaptive std-dev will be updated
        to maintain the target acceptance rate. (e.g: ~30%).
    adaptive_std_factor : :obj:`float` in (0, 1)
        The factor by which the sampler's std-dev is adapted:

        * If acceptance rate is too low → decrease by ``1 - factor``
        * If too high → increase by ``1 + factor``
    **base_sampler_kws
        Additional keyword arguments passed to :class:`~leaspy.samplers.AbstractSampler` init method.
        For example, you may pass `acceptation_history_length`.
    """

    STD_SCALE_FACTOR = 0.5

    def __init__(
        self,
        name: str,
        shape: tuple,
        *,
        n_patients: int,
        scale: Union[float, torch.FloatTensor],
        random_order_dimension: bool = True,
        mean_acceptation_rate_target_bounds: tuple[float, float] = (0.2, 0.4),
        adaptive_std_factor: float = 0.1,
        **base_sampler_kws,
    ):
        super().__init__(
            name,
            shape,
            n_patients=n_patients,
            scale=scale,
            random_order_dimension=random_order_dimension,
            mean_acceptation_rate_target_bounds=mean_acceptation_rate_target_bounds,
            adaptive_std_factor=adaptive_std_factor,
            **base_sampler_kws,
        )

    def validate_scale(self, scale: Union[float, torch.Tensor]) -> torch.Tensor:
        """
        Validate the provided scale.

        Parameters
        ----------
        scale : :obj:`float` or :class:`torch.Tensor`
            The scale to be validated.

        Returns
        -------
        :class:`torch.Tensor` :
            Valid scale.
        """
        scale = super().validate_scale(scale)
        # <!> scale should always be a scalar tensor for individual sampler
        return scale.mean()

    @property
    def shape_adapted_std(self) -> tuple:
        # <!> We do not take into account the dimensionality of
        # individual parameter for acceptation / adaptation
        return (self.n_patients,)

    def _proposed_change(self) -> torch.Tensor:
        """
        Proposal change around the current value with sampler standard deviation.

        .. warning::
            Not to be used for scalar sampling (in population samplers)
            since it would be inefficient!

        Parameters
        ----------
        val : torch.Tensor

        Returns
        -------
        :class:`torch.Tensor` :
            Tensor of shape (self.n_patients, *self.shape).
        """
        std_broadcasting = (slice(None),) + (None,) * self.ndim
        return self.std[std_broadcasting] * torch.randn((self.n_patients, *self.shape))

    def sample(
        self,
        state: State,
        *,
        temperature_inv: float,
    ) -> None:
        """
        Apply a grouped Metropolis-Hastings (MH) sampling step for individual variables.

        For each individual variable:
          - Compute the current attachment and regularity.
          - Propose a new value for the variable.
          - Recompute the attachment and regularity.
          - Accept or reject the proposal using the MH criterion, based on the inverse temperature.


        Parameters
        ----------
        state : :class:`.State`
            Object containing values for all model variables, including latent variables
        temperature_inv : :obj:`float` > 0
            The temperature to use.

        Notes
        -----
        Population variables remain fixed during this step since we are updating individual variables. However:

        - In fit, population variables might have just been updated,
          and their effects not yet propagated to the model.
          In this case, model computations should use the current MCMC state (default behavior).

        - In personalization (mode or mean posterior), population variables are not updated.
          Therefore, the MCMC state should not be used.
        """

        def compute_attachment_regularity():
            # compute neg log-likelihood of just the given variable (tau, xi or sources)
            # (per subject; all dimensions of the individual parameter are summed together)
            return state.get_tensor_values(
                ("nll_attach_ind", f"nll_regul_{self.name}_ind")
            )

        previous_attachment, previous_regularity = compute_attachment_regularity()
        if state["nll_regul_ind_sum_ind"].ndim > 1 :
            nll_regul_ind_sum_ind = state["nll_regul_ind_sum_ind"].value
            nll_cluster = -nll_regul_ind_sum_ind
            probs_ind = torch.nn.Softmax(dim=1)(torch.clamp(nll_cluster, -100.))

        if previous_regularity.ndim == 2 : #it means that we have clusters and for the individual parameters we calculate a regularity term per cluster
            previous_regularity = (probs_ind * previous_regularity).sum(dim=1)

        # with state.auto_fork():
        state.put(
            self.name,
            self._proposed_change(),
            accumulate=True,  # out-of-place addition
        )

        # alpha is per individual and > 0, shape = (n_individuals,)
        # if new is "better" than previous, then alpha > 1 so it will always be accepted in `_group_metropolis_step`
        new_attachment, new_regularity = compute_attachment_regularity()

        if state["nll_regul_ind_sum_ind"].ndim > 1:
            nll_regul_ind_sum_ind = state["nll_regul_ind_sum_ind"].value
            nll_cluster = -nll_regul_ind_sum_ind
            probs_ind = torch.nn.Softmax(dim=1)(torch.clamp(nll_cluster, -100.))

        if new_regularity.ndim == 2:  # it means that we have clusters and for the individual parameters we calculate a regularity term per cluster
            new_regularity = (probs_ind * new_regularity).sum(dim=1)
        alpha = torch.exp(
            -1
            * (
                (new_regularity - previous_regularity) * temperature_inv
                + (new_attachment - previous_attachment)
            )
        )
        accepted = self._group_metropolis_step(alpha)

        # Partial reversion where proposals were not accepted
        state.revert(~accepted)

        self._update_acceptation_rate(accepted.float())
        self._update_std()



from leaspy.io.data.dataset import Dataset
from leaspy.utils.weighted_tensor import WeightedTensor
from leaspy.variables.distributions import (
    WeibullRightCensored,
    WeibullRightCensoredWithSources,
)
from leaspy.variables.specs import (
    LinkedVariable,
    VariableInterface,
    VariableName,
)

from ._base import ObservationModel

__all__ = [
    "AbstractWeibullRightCensoredObservationModel",
    "WeibullRightCensoredObservationModel",
    "WeibullRightCensoredWithSourcesObservationModel",
]


class AbstractWeibullRightCensoredObservationModel(ObservationModel):
    @staticmethod
    def getter(dataset: Dataset) -> WeightedTensor:
        if dataset.event_time is None or dataset.event_bool is None:
            raise ValueError(
                "Provided dataset is not valid. "
                "Both values and mask should be not None."
            )
        return WeightedTensor(dataset.event_time, dataset.event_bool)

    def get_variables_specs(
        self,
        named_attach_vars: bool = True,
    ) -> dict[VariableName, VariableInterface]:
        """Automatic specifications of variables for this observation model."""

        specs = super().get_variables_specs(named_attach_vars)

        specs[f"predictions_{self.name}"] = LinkedVariable(
            self.dist.get_func("compute_predictions", self.name)
        )

        return specs


class WeibullRightCensoredObservationModel(
    AbstractWeibullRightCensoredObservationModel
):
    string_for_json = "weibull-right-censored"

    def __init__(
        self,
        nu: VariableName,
        rho: VariableName,
        xi: VariableName,
        tau: VariableName,
        **extra_vars: VariableInterface,
    ):
        super().__init__(
            name="event",
            getter=self.getter,
            dist=WeibullRightCensored(nu, rho, xi, tau),
            extra_vars=extra_vars,
        )

    @classmethod
    def default_init(self, **kwargs):
        return self(
            nu=kwargs.pop("nu", "nu"),
            rho=kwargs.pop("rho", "rho"),
            xi=kwargs.pop("xi", "xi"),
            tau=kwargs.pop("tau", "tau"),
        )


class WeibullRightCensoredWithSourcesObservationModel(
    AbstractWeibullRightCensoredObservationModel
):
    string_for_json = "weibull-right-censored-with-sources"

    def __init__(
        self,
        nu: VariableName,
        rho: VariableName,
        xi: VariableName,
        tau: VariableName,
        survival_shifts: VariableName,
        **extra_vars: VariableInterface,
    ):
        super().__init__(
            name="event",
            getter=self.getter,
            dist=WeibullRightCensoredWithSources(nu, rho, xi, tau, survival_shifts),
            extra_vars=extra_vars,
        )

    @classmethod
    def default_init(self, **kwargs):
        return self(
            nu=kwargs.pop("nu", "nu"),
            rho=kwargs.pop("rho", "rho"),
            xi=kwargs.pop("xi", "xi"),
            tau=kwargs.pop("tau", "tau"),
            survival_shifts=kwargs.pop("survival_shifts", "survival_shifts"),
        )



import warnings
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
import torch

from leaspy.algo import AlgorithmName, AlgorithmSettings
from leaspy.exceptions import LeaspyModelInputError
from leaspy.io.data.dataset import Data, Dataset
from leaspy.io.outputs import IndividualParameters
from leaspy.io.outputs.result import Result
from leaspy.utils.typing import DictParamsTorch, FeatureType, IDType, KwargsType

__all__ = [
    "InitializationMethod",
    "BaseModel",
    "ModelInterface",
]


class InitializationMethod(str, Enum):
    """Possible initialization methods for Leaspy models.

    Attributes
    ----------
    DEFAULT : :obj:`str`
        Default initialization method.

    RANDOM : :obj:`str`
        Random initialization method.
    """

    DEFAULT = "default"
    RANDOM = "random"


class ModelInterface(ABC):
    """This is the public interface for Leaspy models.
    It defines the methods and properties that all models should implement.
    It is not meant to be instantiated directly, but rather to be inherited by concrete model classes.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the model.

        Returns
        -------
        :obj:`str`

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def is_initialized(self) -> bool:
        """True if the model is initialized, False otherwise.

        Returns
        -------
        :obj:`bool`

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def dimension(self) -> int:
        """Number of features.

        Returns
        -------
        :obj:`int`

        Raises
        ------
        NotImplementedError"""
        raise NotImplementedError

    @property
    @abstractmethod
    def features(self) -> list[FeatureType]:
        """List of model features (`None` if not initialization).

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def parameters(self) -> DictParamsTorch:
        """Dictionary of values for model parameters.

        Returns
        -------
        :class:`~leaspy.utils.typing.DictParamsTorch`

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def hyperparameters(self) -> DictParamsTorch:
        """Dictionary of values for model hyperparameters.

        Returns
        -------
        :class:`~leaspy.utils.typing.DictParamsTorch`

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    def save(self, path: Union[str, Path], **kwargs) -> None:
        """Save model as json model parameter file.

        Parameters
        ----------
        path : :obj:`str` or :obj:`Path`
            The path to store the model's parameters.

        **kwargs : :obj:`dict`
            Additional parameters for writing.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def load(cls, path_to_model_settings: Union[str, Path]):
        """Load a model from a json model parameter file.

        Parameters
        ----------
        path_to_model_settings : :obj:`str` or :obj:`Path`
            The path to the model's parameters file.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    def fit(
        self,
        data: Optional[Union[pd.DataFrame, Data, Dataset]] = None,
        algorithm: Optional[Union[str, AlgorithmName]] = None,
        algorithm_settings: Optional[AlgorithmSettings] = None,
        algorithm_settings_path: Optional[Union[str, Path]] = None,
        **kwargs,
    ):
        r"""Estimate the model's parameters :math:`\theta` for a given dataset and a given algorithm.

        These model's parameters correspond to the fixed-effects of the mixed-effects model.

        There are three ways to provide parameters to the fitting algorithm:

        1. By providing an instance of :class:`~leaspy.algo.AlgorithmSettings`
        2. By providing a path to a serialized :class:`~leaspy.algo.AlgorithmSettings`
        3. By providing the algorithm name and parameters directly

        If settings are provided in multiple ways, the order above will prevail.

        Parameters
        ----------
        data : :obj:`pd.DataFrame` or :class:`~leaspy.io.Data` or :class:`~leaspy.io.Dataset`, optional
            Contains the information of the individuals, in particular the time-points
            :math:`(t_{i,j})` and the observations :math:`(y_{i,j})`.

        algorithm : :obj:`str` or :class:`~leaspy.algo.base.AlgorithmName`,optional
            The name of the algorithm to use.

            .. note::
                Use this if you want to provide algorithm settings through kwargs.

        algorithm_settings : :class:`~leaspy.algo.AlgorithmSettings`, optional
            The algorithm settings to use.

            .. note::
                Use this if you want to customize algorithm settings through the
                :class:`~leaspy.algo.AlgorithmSettings` class.
                If provided, the fit will rely on these settings.

        algorithm_settings_path : :obj:`str` or :obj:`Path`, optional
            The path to the algorithm settings file.

            .. note::
                If provided, the settings from the file will be used instead of the
                settings provided through kwarsg.

        **kwargs : :obj:`dict`
            Contains the algorithm's settings.

        Raises
        ------
        NotImplementedError

        Examples
        --------
        Fit a logistic model on a longitudinal dataset, display the group parameters

        >>> from leaspy.models import LogisticModel
        >>> from leaspy.datasets import load_dataset
        >>> putamen_df = load_dataset("parkinson-putamen")
        >>> model = LogisticModel(name="test-model-logistic")
        >>> model.fit(putamen_df, "mcmc_saem", seed=0, print_periodicity=50)
        >>> print(model)
        === MODEL ===
        betas_mean : []
        log_g_mean : [-0.8394]
        log_v0_mean : [-3.7930]
        noise_std : 0.021183
        tau_mean : [64.6920]
        tau_std : [10.0864]
        xi_std : [0.5232]
        """
        raise NotImplementedError

    @abstractmethod
    def personalize(
        self,
        data: Optional[Union[pd.DataFrame, Data, Dataset]] = None,
        algorithm: Optional[Union[str, AlgorithmName]] = None,
        algorithm_settings: Optional[AlgorithmSettings] = None,
        algorithm_settings_path: Optional[Union[str, Path]] = None,
        **kwargs,
    ) -> IndividualParameters:
        r"""Estimate individual parameters for each `ID` of a given dataset.

        These individual parameters correspond to the random-effects :math:`(z_{i,j})` of the mixed-effects model.

        Parameters
        ----------
        data : :obj:`pd.DataFrame` or :class:`~leaspy.io.Data` or :class:`~leaspy.io.Dataset`, optional
            Contains the information of the individuals, in particular the time-points
            :math:`(t_{i,j})` and the observations :math:`(y_{i,j})`.

        algorithm : :obj:`str` or :class:`~leaspy.algo.base.AlgorithmName`, optional
            The name of the algorithm to use.

        algorithm_settings : :class:`~leaspy.algo.AlgorithmSettings`, optional
            The algorithm settings to use.

            .. note::
                Use this if you want to customize algorithm settings through the
                :class:`~leaspy.algo.AlgorithmSettings` class.
                If provided, the fit will rely on these settings.

        algorithm_settings_path : :obj:`str` or :obj:`Path`, optional
            The path to the algorithm settings file.

            .. note::
                If provided, the settings from the file will be used instead of the settings provided.

        **kwargs : :obj:`dict`
            Contains the algorithm's settings.

        Returns
        -------
        ips : :class:`~leaspy.io.IndividualParameters`
            Individual parameters computed.

        Raises
        ------
        NotImplementedError

        Examples
        --------
        Compute the individual parameters for a given longitudinal dataset and calibrated model, then
        display the histogram of the log-acceleration:

        >>> from leaspy.datasets import load_model, load_dataset
        >>> model = load_model("parkinson-putamen")
        >>> putamen_df = load_dataset("parkinson-putamen")
        >>> individual_parameters = model.personalize(putamen_df, "scipy_minimize", seed=0)
        """
        raise NotImplementedError

    @abstractmethod
    def estimate(
        self,
        timepoints: Union[pd.MultiIndex, dict[IDType, list[float]]],
        individual_parameters: IndividualParameters,
        *,
        to_dataframe: Optional[bool] = None,
    ) -> Union[pd.DataFrame, dict[IDType, np.ndarray]]:
        r"""Return the model values for individuals characterized by their individual parameters :math:`z_i` at time-points :math:`(t_{i,j})_j`.

        Parameters
        ----------
        timepoints : :obj:`pd.MultiIndex` or :obj:`dict` [:obj:`IDType`, :obj:`list` [:obj:`float` ] ]
            Contains, for each individual, the time-points to estimate.
            It can be a unique time-point or a list of time-points.

        individual_parameters : :class:`~leaspy.io.IndividualParameters`
            Corresponds to the individual parameters of individuals.

        to_dataframe : :obj:`bool`, optional
            Whether to output a dataframe of estimations?
            If None: default is to be True if and only if timepoints is a `pandas.MultiIndex`

        Returns
        -------
        individual_trajectory : :obj:`pd.MultiIndex` or :obj:`dict` [:obj:`IDType`, :obj:`list` [:obj:`float`]]
            Key: patient indices.
            Value: :class:`numpy.ndarray` of the estimated value, in the shape (number of timepoints, number of features)

        Raises
        ------
        NotImplementedError

        Examples
        --------
        Given the individual parameters of two subjects, estimate the features of the first
        at 70, 74 and 80 years old and at 71 and 72 years old for the second.

        >>> from leaspy.datasets import load_model, load_individual_parameters, load_dataset
        >>> model = load_model("parkinson-putamen")
        >>> individual_parameters = load_individual_parameters("parkinson-putamen")
        >>> df_train = load_dataset("parkinson-putamen-train_and_test").xs("train", level="SPLIT")
        >>> timepoints = {'GS-001': (70, 74, 80), 'GS-002': (71, 72)}
        >>> estimations = model.estimate(timepoints, individual_parameters)
        """
        raise NotImplementedError

    @abstractmethod
    def simulate(
        self,
        individual_parameters: IndividualParameters,
        data: Optional[Union[pd.DataFrame, Data, Dataset]] = None,
        **kwargs,
    ):
        """Run the simulation pipeline using a leaspy model.
         This method simulates longitudinal data using the given leaspy model.
         It performs the following steps:
         - Retrieves individual parameters (IP) from fixed effects of the model.
         - Loads the specified Leaspy model.
         - Generates visit ages (timepoints) for each individual (based on specifications
           in visits_type).
         - Simulates observations at those visit ages.
         - Packages the result into a `Result` object, including simulated data,
           individual parameters, and the model's noise standard deviation.

         Parameters
         ----------
         individual_parameters : :class:`~leaspy.io.IndividualParameters`
             Individual parameters to use for the simulation.

         data : :obj:`pd.DataFrame` or :class:`~leaspy.io.Data` or :class:`~leaspy.io.Dataset`
             Data object. If None, returns empty Result.

         **kwargs : :obj:`dict`
         raise NotImplementedError
        `  Additional arguments for algorithm settings.

         Returns
         -------
         simulated_data : :class:`~leaspy.io.outputs.result.Result`
             Contains the generated individual parameters & the corresponding generated scores.
             Returns empty Result if any required input is None.

         Raises
         ------
         NotImplementedError
        """
        raise NotImplementedError


class BaseModel(ModelInterface):
    """Base model class from which all ``Leaspy`` models should inherit.

    It implements the :class:`~leaspy.models.ModelInterface`.
    """

    def __init__(self, name: str, **kwargs):
        self._is_initialized: bool = False
        self._name = name
        user_provided_dimension, user_provided_features = (
            self._validate_user_provided_dimension_and_features_at_init(**kwargs)
        )
        self._features: Optional[list[FeatureType]] = user_provided_features
        self._dimension: Optional[int] = user_provided_dimension
        self.initialization_method: InitializationMethod = InitializationMethod.DEFAULT
        if "initialization_method" in kwargs:
            self.initialization_method = InitializationMethod(
                kwargs["initialization_method"]
            )

    @property
    def name(self) -> str:
        """The name of the model.

        Returns
        -------
        :obj:`str`
            The name of the model.
        """

        return self._name

    @property
    def is_initialized(self) -> bool:
        """True if the model is initialized, False otherwise.

        Returns
        -------
        :obj:`bool`
            True if the model is initialized, False otherwise.
        """
        return self._is_initialized

    def _validate_user_provided_dimension_and_features_at_init(
        self,
        **kwargs,
    ) -> tuple[Optional[int], Optional[list[FeatureType]]]:
        """Validate user provided dimension and features at model initialization.

        Parameters
        ----------
        **kwargs : :obj:`dict`
            Keyword arguments that may contain 'features' and 'dimension'.

        Returns
        -------
        :obj:`tuple` [:obj:`int`, :obj:`list` [:obj:`FeatureType`]]
            A tuple containing the validated dimension and features.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If the provided dimension is not an integer or if the features are not sizeable.
            If the provided dimension does not match the number of features.
        """
        from collections.abc import Sized

        user_provided_features = kwargs.pop("features", None)
        user_provided_dimension = kwargs.pop("dimension", None)
        if user_provided_dimension is not None and not isinstance(
            user_provided_dimension, int
        ):
            raise LeaspyModelInputError(
                f"{self.__class__.__name__} model '{self.name}' cannot be instantiated with "
                f"dimension = {user_provided_dimension}, of type {type(user_provided_dimension)}. "
                "The number of dimension must be an integer."
            )
        if user_provided_features is not None and not isinstance(
            user_provided_features, Sized
        ):
            raise LeaspyModelInputError(
                f"{self.__class__.__name__} model '{self.name}' cannot be instantiated with "
                f"features = {user_provided_features}. The model's features must be a sizeable object."
            )
        if (
            user_provided_features is not None
            and user_provided_dimension is not None
            and user_provided_dimension != len(user_provided_features)
        ):
            raise LeaspyModelInputError(
                f"{self.__class__.__name__} model '{self.name}' cannot be instantiated with "
                f"dimension = {user_provided_dimension} and features = {user_provided_features}. "
                "The model dimension must match the number of features."
            )
        return user_provided_dimension, user_provided_features

    @property
    def features(self) -> Optional[list[FeatureType]]:
        """List of model features (`None` if not initialization).

        Returns
        -------
        : :obj:`list` [:obj:`FeatureType`], optional
            The features of the model, or None if not initialized.
        """
        return self._features

    @features.setter
    def features(self, features: Optional[list[FeatureType]]):
        """Set the model's features.

        Parameters
        ----------
        features : :obj:`list`[:obj:`FeatureType`], optional
            The features to set for the model. If None, it resets the features.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If the provided features do not match the model's dimension.
        """
        if features is None:
            # used to reset features
            self._features = None
            return

        if self.dimension is not None and len(features) != self.dimension:
            raise LeaspyModelInputError(
                f"Cannot set the model's features to {features}, because "
                f"the model has been configured with a dimension of {self.dimension}."
            )
        self._features = features

    @property
    def dimension(self) -> Optional[int]:
        """Number of features.

        Returns
        -------
        :obj:`int`, optional
            The dimension of the model, or None if not initialized.
        """
        if self._dimension is not None:
            return self._dimension
        if self.features is not None:
            return len(self.features)
        return None

    @dimension.setter
    def dimension(self, dimension: int):
        """Set the model's dimension.

        Parameters
        ----------
        dimension : :obj:`int`
            The dimension to set for the model.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If the model has already been initialized with features that do not match the new dimension.
        """
        if self.features is None:
            self._dimension = dimension
        elif len(self.features) != dimension:
            raise LeaspyModelInputError(
                f"Model has {len(self.features)} features. Cannot set the dimension to {dimension}."
            )

    def _validate_compatibility_of_dataset(
        self, dataset: Optional[Dataset] = None
    ) -> None:
        """Raise if the given Dataset is not compatible with the current model.

        Parameters
        ----------
        dataset : :class:`~leaspy.io.data.dataset.Dataset`, optional
            The class we want to model.

        Raises
        ------
        :exc:`.LeaspyModelInputError` :
            - If the Dataset has a number of dimensions smaller than 2.
            - If the Dataset does not have the same dimensionality as the model.
            - If the Dataset's headers do not match the model's.
        """
        if not dataset:
            return
        if self.dimension is not None and dataset.dimension != self.dimension:
            raise LeaspyModelInputError(
                f"Unmatched dimensions: {self.dimension} (model) ≠ {dataset.dimension} (data)."
            )
        if self.features is not None and dataset.headers != self.features:
            raise LeaspyModelInputError(
                f"Unmatched features: {self.features} (model) ≠ {dataset.headers} (data)."
            )

    def initialize(self, dataset: Optional[Dataset] = None) -> None:
        """Initialize the model given a :class:`.Dataset` and an initialization method.

        After calling this method :attr:`is_initialized` should be ``True`` and model
        should be ready for use.

        Parameters
        ----------
        dataset : :class:`~leaspy.io.data.dataset.Dataset`, optional
            The dataset we want to initialize from.
        """
        if self.is_initialized and self.features is not None:
            # we also test that self.features is not None, since for `ConstantModel`:
            # `is_initialized`` is True but as a mock for being personalization-ready,
            # without really being initialized!
            warn_msg = "<!> Re-initializing an already initialized model."
            if dataset and dataset.headers != self.features:
                warn_msg += (
                    f" Overwriting previous model features ({self.features}) "
                    f"with new ones ({dataset.headers})."
                )
                # wait validation of compatibility to store new features
                self.features = None
            warnings.warn(warn_msg)
        self._validate_compatibility_of_dataset(dataset)
        self.features = dataset.headers if dataset else None
        self._is_initialized = True

    def save(self, path: Union[str, Path], **kwargs) -> None:
        """Save model as json model parameter file.

        Parameters
        ----------
        path : :obj:`str` or :obj:`Path`
            The path to store the model's parameters.

        **kwargs : :obj:`dict`
            Additional parameters for writing.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If the model is not initialized.
        """
        import json
        from inspect import signature

        export_kws = {
            k: kwargs.pop(k) for k in signature(self.to_dict).parameters if k in kwargs
        }
        model_settings = self.to_dict(**export_kws)
        kwargs = {"indent": 2, **kwargs}
        with open(path, "w") as fp:
            json.dump(model_settings, fp, **kwargs)

    def to_dict(self, **kwargs) -> KwargsType:
        """Export model as a dictionary ready for export.

        Returns
        -------
        :obj:`KwargsType`
            The model instance serialized as a dictionary.
        """
        from leaspy import __version__

        from .utilities import tensor_to_list

        return {
            "leaspy_version": __version__,
            "name": self.name,
            "features": self.features,
            "dimension": self.dimension,
            "hyperparameters": {
                k: tensor_to_list(v) for k, v in (self.hyperparameters or {}).items()
            },
            "parameters": {
                k: tensor_to_list(v) for k, v in (self.parameters or {}).items()
            },
        }

    @classmethod
    def load(cls, path_to_model_settings: Union[str, Path]):
        """Load a model from a json model parameter file.

        Parameters
        ----------
        path_to_model_settings : :obj:`str` or :obj:`Path`
            The path to the model's parameters file.

        Returns
        -------
        :class:`~leasp.models.base.BaseModel`
            An instance of the model loaded from the file.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If the model settings file is not found or cannot be read.
        """

        from .factory import model_factory
        from .settings import ModelSettings

        reader = ModelSettings(path_to_model_settings)
        instance = model_factory(reader.name, **reader.hyperparameters)
        instance.load_parameters(reader.parameters)
        instance._is_initialized = True
        return instance

    @abstractmethod
    def load_parameters(self, parameters: KwargsType) -> None:
        """Load model parameters from a dictionary.

        Parameters
        ----------
        parameters : :obj:`KwargsType`
            The parameters to load into the model.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    def fit(
        self,
        data: Optional[Union[pd.DataFrame, Data, Dataset]] = None,
        algorithm: Optional[Union[str, AlgorithmName]] = None,
        algorithm_settings: Optional[AlgorithmSettings] = None,
        algorithm_settings_path: Optional[Union[str, Path]] = None,
        **kwargs,
    ):
        """Estimate the model's parameters for a given dataset and a given algorithm.
        These model's parameters correspond to the fixed-effects of the mixed-effects model.

        Parameters
        ----------
        data : :obj:`pd.DataFrame` or :class:`~leaspy.io.Data` or :class:`~leaspy.io.Dataset`, optional
            Contains the information of the individuals, in particular the time-points
            :math:`(t_{i,j})` and the observations :math:`(y_{i,j})`.
        algorithm : :obj:`str` or :class:`~leaspy.algo.base.AlgorithmName`, optional
            The name of the algorithm to use.
            Use this if you want to provide algorithm settings through kwargs.
        algorithm_settings : :class:`~leaspy.algo.AlgorithmSettings`, optional
            The algorithm settings to use.
            Use this if you want to customize algorithm settings through the
            :class:`~leaspy.algo.AlgorithmSettings` class.
            If provided, the fit will rely on these settings.
        algorithm_settings_path : :obj:`str` or :obj:`Path`, optional
            The path to the algorithm settings file.
            If provided, the settings from the file will be used instead of the settings provided.
        **kwargs : :obj:`dict`
            Contains the algorithm's settings.
        """
        if (dataset := BaseModel._get_dataset(data)) is None:
            return
        if not self.is_initialized:
            self.initialize(dataset)
        if (
            algorithm := BaseModel._get_algorithm(
                algorithm, algorithm_settings, algorithm_settings_path, **kwargs
            )
        ) is None:
            return
        algorithm.run(self, dataset)

    @staticmethod
    def _get_dataset(
        data: Optional[Union[pd.DataFrame, Data, Dataset]] = None,
    ) -> Optional[Dataset]:
        """Process user provided data and return a Dataset object.

        Parameters
        ----------
        data : :obj:`pd.DataFrame` or :class:`~leaspy.io.Data` or :class:`~leaspy.io.Dataset`, optional
            The data to process. If None, returns None.

        Returns
        -------
        :class:`~leaspy.io.data.dataset.Dataset`, optional
            A Dataset object if data is provided and valid, otherwise None.
        """
        if data is None:
            return None
        if isinstance(data, pd.DataFrame):
            data = Data.from_dataframe(data)
        return Dataset(data) if isinstance(data, Data) else data

    @staticmethod
    def _get_algorithm(
        algorithm: Optional[Union[str, AlgorithmName]] = None,
        algorithm_settings: Optional[AlgorithmSettings] = None,
        algorithm_settings_path: Optional[Union[str, Path]] = None,
        **kwargs,
    ):
        """Process user provided algorithm and return the corresponding algorithm instance.

        Parameters
        ----------
        algorithm : :obj:`str` or :class:`~leaspy.algo.base.AlgorithmName`, optional
            The name of the algorithm to use. If None, returns None.
        algorithm_settings : :class:`~leaspy.algo.AlgorithmSettings`, optional
            The algorithm settings to use. If None, returns None.
        algorithm_settings_path : :obj:`str` or :obj:`Path`, optional
            The path to the algorithm settings file. If None, returns None.
        **kwargs : :obj:`dict`
            Additional parameters for the algorithm settings.

        Returns
        -------
        :class:`~leaspy.algo.base.AlgorithmInterface`, optional
            An instance of the algorithm if provided, otherwise None."""
        from leaspy.algo import AlgorithmName, AlgorithmSettings, algorithm_factory

        if algorithm_settings is not None:
            settings = algorithm_settings
        elif algorithm_settings_path is not None:
            settings = AlgorithmSettings.load(algorithm_settings_path)
        else:
            algorithm = AlgorithmName(algorithm) if algorithm else None
            if algorithm is None:
                return None
            settings = AlgorithmSettings(algorithm.value, **kwargs)
            settings.set_logs(**kwargs)
        return algorithm_factory(settings)

    def personalize(
        self,
        data: Optional[Union[pd.DataFrame, Data, Dataset]] = None,
        algorithm: Optional[Union[str, AlgorithmName]] = None,
        algorithm_settings: Optional[AlgorithmSettings] = None,
        algorithm_settings_path: Optional[Union[str, Path]] = None,
        **kwargs,
    ) -> IndividualParameters:
        """Estimate individual parameters for each `ID` of a given dataset.
        These individual parameters correspond to the random-effects :math:`(z_{i,j})` of the mixed-effects model.

        Parameters
        ----------
        data : :obj:`pd.DataFrame` or :class:`~leaspy.io.Data` or :class:`~leaspy.io.Dataset`, optional
            Contains the information of the individuals, in particular the time-points
            :math:`(t_{i,j})` and the observations :math:`(y_{i,j})`.
        algorithm : :obj:`str` or :class:`~leaspy.algo.base.AlgorithmName`, optional
            The name of the algorithm to use.
        algorithm_settings : :class:`~leaspy.algo.AlgorithmSettings`, optional
            The algorithm settings to use.
            Use this if you want to customize algorithm settings through the
            :class:`~leaspy.algo.AlgorithmSettings` class.
            If provided, the fit will rely on these settings.
        algorithm_settings_path : :obj:`str` or :obj:`Path`, optional
            The path to the algorithm settings file.
            If provided, the settings from the file will be used instead of the settings provided.
        **kwargs : :obj:`dict`
            Contains the algorithm's settings.

        Returns
        -------
        :class:`~leaspy.io.outputs.IndividualParameters`
            Individual parameters computed.
        """

        from leaspy.exceptions import LeaspyInputError

        if not self.is_initialized:
            raise LeaspyInputError("Model has not been initialized")
        if (dataset := BaseModel._get_dataset(data)) is None:
            return IndividualParameters()
        if (
            algorithm := BaseModel._get_algorithm(
                algorithm, algorithm_settings, algorithm_settings_path, **kwargs
            )
        ) is None:
            return IndividualParameters()
        return algorithm.run(self, dataset)

    def estimate(
        self,
        timepoints: Union[pd.MultiIndex, dict[IDType, list[float]]],
        individual_parameters: IndividualParameters,
        *,
        to_dataframe: Optional[bool] = None,
    ) -> Union[pd.DataFrame, dict[IDType, np.ndarray]]:
        """Return the model values for individuals characterized by their individual parameters :math:`z_i` at time-points :math:`(t_{i,j})_j`.

        Parameters
        ----------
        timepoints : :obj:`pd.MultiIndex` or :obj:`dict` [:obj:`IDType`, :obj:`list` [:obj:`float`]]
            Contains, for each individual, the time-points to estimate.
            It can be a unique time-point or a list of time-points.

        individual_parameters : :class:`~leaspy.io.IndividualParameters`
            Corresponds to the individual parameters of individuals.

        to_dataframe : :obj:`bool`, optional
            Whether to output a dataframe of estimations?
            If None: default is to be True if and only if timepoints is a `pandas.MultiIndex`

        Returns
        -------
        individual_trajectory : :obj:`pd.DataFrame` or :obj:`dict` [:obj:`IDType`, :obj:`np.ndarray`]
            Key: patient indices.
            Value: :class:`numpy.ndarray` of the estimated value, in the shape (number of timepoints, number of features)
        """

        estimations = {}
        ix = None
        # get timepoints to estimate from index
        if isinstance(timepoints, pd.MultiIndex):
            # default output is pd.DataFrame when input as pd.MultiIndex
            if to_dataframe is None:
                to_dataframe = True
            ix = timepoints
            timepoints = {
                subj_id: tpts.values
                for subj_id, tpts in timepoints.to_frame()["TIME"].groupby("ID")
            }
        for subj_id, tpts in timepoints.items():
            ip = individual_parameters[subj_id]
            est = self.compute_individual_trajectory(tpts, ip).cpu().numpy()
            # 1 individual at a time --> squeeze the first dimension of the array
            estimations[subj_id] = est[0]

        # convert to proper dataframe
        if to_dataframe:
            estimations = pd.concat(
                {
                    subj_id: pd.DataFrame(  # columns names may be directly embedded in the dictionary after a `postprocess_model_estimation`
                        ests,
                        columns=None if isinstance(ests, dict) else self.features,
                        index=timepoints[subj_id],
                    )
                    for subj_id, ests in estimations.items()
                },
                names=["ID", "TIME"],
            )
            # reindex back to given index being careful to index order (join so to handle multi-levels cases)
            if ix is not None:
                # we need to explicitly pass `on` to preserve order of index levels
                # and to explicitly pass columns to preserve 2D columns when they are
                empty_df_like_ests = pd.DataFrame(
                    [], index=ix, columns=estimations.columns
                )
                estimations = empty_df_like_ests[[]].join(
                    estimations, on=["ID", "TIME"]
                )

        return estimations

    @abstractmethod
    def compute_individual_trajectory(
        self,
        timepoints: list[float],
        individual_parameters: IndividualParameters,
    ) -> torch.Tensor:
        """Compute the model values for an individual characterized by their individual parameters at given time-points.
        Parameters
        ----------
        timepoints : :obj:`list` [:obj:`float`]
            The time-points at which to compute the model values.
        individual_parameters : :class:`~leaspy.io.IndividualParameters`
            The individual parameters of the individual.

        Returns
        -------
        :class:`torch.Tensor`
            The computed model values for the individual at the given time-points.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in the subclass.
        """
        raise NotImplementedError

    def simulate(
        self,
        algorithm: Optional[Union[str, AlgorithmName]] = None,
        algorithm_settings: Optional[AlgorithmSettings] = None,
        algorithm_settings_path: Optional[Union[str, Path]] = None,
        **kwargs,
    ):
        """
        Run the simulation pipeline using a leaspy model.

        This method simulates longitudinal data using the given leaspy model.
        It performs the following steps:
        - Retrieves individual parameters (IP) from fixed effects of the model.
        - Loads the specified Leaspy model.
        - Generates visit ages (timepoints) for each individual (based on specifications
        in visits_type)
        - Simulates observations at those visit ages.
        - Packages the result into a `Result` object, including simulated data,
        individual parameters, and the model's noise standard deviation.

        Parameters
        ----------
        algorithm : :obj:`str` or :class:`~leaspy.algo.base.AlgorithmName`, optional
            The name of the algorithm to use.
            Use this if you want to provide algorithm settings through kwargs.
        algorithm_settings : :class:`~leaspy.algo.AlgorithmSettings`, optional
            The algorithm settings to use.
            Use this if you want to customize algorithm settings through the
            :class:`~leaspy.algo.AlgorithmSettings` class.
            If provided, the fit will rely on these settings.
        algorithm_settings_path : :obj:`str` or :obj:`Path`, optional
            The path to the algorithm settings file.
            If provided, the settings from the file will be used instead of the settings provided.
        **kwargs : :obj:`dict`
            Contains the algorithm's settings.

        Returns
        -------
        simulated_data : :class:`~leaspy.io.outputs.result.Result`
            Contains the generated individual parameters & the corresponding generated scores.
            Returns empty Result if any required input is None.


        Notes
        -----
        To generate a new subject, first we estimate the joined distribution of the individual parameters and the
        reparametrized baseline ages. Then, we randomly pick a new point from this distribution, which define the
        individual parameters & baseline age of our new subjects. Then, we generate the timepoints
        following the baseline age. Then, from the model and the generated timepoints and individual parameters, we
        compute the corresponding values estimations. Then, we add some noise to these estimations, which is the
        same noise-model as the one from your model by default. But, you may customize it by setting the `noise` keyword.

        Examples
        --------
        Use a calibrated model & individual parameters to simulate new subjects similar to the ones you have:

        >>> from leaspy.models import LogisticModel
        >>> from leaspy.io.data import Data
        >>> from leaspy.datasets import load_dataset, load_leaspy_instance, load_individual_parameters
        >>> putamen_df = load_dataset("parkinson-putamen-train_and_test")
        >>> data = Data.from_dataframe(putamen_df.xs('train', level='SPLIT'))
        >>> leaspy_logistic = load_leaspy_instance("parkinson-putamen-train")
        >>> visits_params = {'patient_number':200,
                 'visit_type': "random",
                 'first_visit_mean' : 0.,
                 'first_visit_std' : 0.4,
                 'time_follow_up_mean' : 11,
                 'time_follow_up_std' : 0.5,
                 'distance_visit_mean' : 2/12,
                 'distance_visit_std' : 0.75/12,
                 'min_spacing_between_visits': 1/365
                }
        >>> simulated_data = model.simulate( algorithm="simulate", features=["MDS1_total", "MDS2_total", "MDS3_off_total", 'SCOPA_total','MOCA_total','REM_total','PUTAMEN_R','PUTAMEN_L','CAUDATE_R','CAUDATE_L'],visit_parameters= visits_params  )
        """
        from leaspy.exceptions import LeaspyInputError

        if not self.is_initialized:
            raise LeaspyInputError("Model has not been initialized")

        if (
            algorithm := BaseModel._get_algorithm(
                algorithm, algorithm_settings, algorithm_settings_path, **kwargs
            )
        ) is None:
            # if no algorithm is provided, we cannot simulate anything
            return Result()

        return algorithm.run(self)

    def __str__(self) -> str:
        """String representation of the model.
        Returns
        -------
        :obj:`str`
            A string representation of the model, including its name, dimension, features, and parameters.
        """
        from .utilities import serialize_tensor

        output = f"=== {self.__class__.__name__} {self.name} ==="
        output += f"\ndimension : {self.dimension}\nfeatures : {self.features}"
        output += serialize_tensor(self.parameters)

        # TODO/WIP obs models...
        # nm_props = export_noise_model(self.noise_model)
        # nm_name = nm_props.pop('name')
        # output += f"\nnoise-model : {nm_name}"
        # output += self._serialize_tensor(nm_props, indent="  ")

        return output
